"""Wraps the ``json`` module, so that own classes get encoded."""

import json
from dataclasses import asdict
from typing import Any
from uuid import UUID

from .entry import Entry
from .queue import Queue
from .result import Result


class SyngEncoder(json.JSONEncoder):
    """
    Encoder of :py:class:`Entry`, :py:class`Queue`, :py:class`Result` and UUID.

    Entry and Result are ``dataclasses``, so they are mapped to their
    dictionary representation.

    UUID is repersented by its string, and Queue will be represented by a list.
    """

    def default(self, o: Any) -> Any:
        """Implement the encoding."""
        if isinstance(o, Entry):
            return asdict(o)
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, Result):
            return asdict(o)
        if isinstance(o, Queue):
            return o.to_list()
        return json.JSONEncoder.default(self, o)


def dumps(obj: Any, **kw: Any) -> str:
    """Wrap around ``json.dumps`` with the :py:class:`SyngEncoder`."""
    return json.dumps(obj, cls=SyngEncoder, **kw)


def dump(obj: Any, fp: Any, **kw: Any) -> None:
    """Forward everything to ``json.dump``."""
    json.dump(obj, fp, cls=SyngEncoder, **kw)


def loads(string: str, **kw: Any) -> Any:
    """Forward everything to ``json.loads``."""
    return json.loads(string, **kw)


def load(fp: Any, **kw: Any) -> Any:
    """Forward everything to ``json.load``."""
    return json.load(fp, **kw)



from importlib.metadata import version, PackageNotFoundError
from packaging.version import Version

try:
    __version__ = version("syng")
    SYNG_VERSION = Version(__version__).release
except PackageNotFoundError:
    __version__ = "unknown"
    SYNG_VERSION = (0, 0, 0)


SYNG_PROTOCOL_VERSION = (2, 2, 0)



from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


class Option(Generic[T]):
    pass


@dataclass
class ConfigOption(Generic[T]):
    type: Option[T]
    description: str
    default: T
    send_to_server: bool = False


class BoolOption(Option[bool]):
    pass


class IntOption(Option[int]):
    pass


class StrOption(Option[str]):
    pass


class PasswordOption(Option[str]):
    pass


class FolderOption(Option[str]):
    pass


class FileOption(Option[str]):
    pass


class ListStrOption(Option[list[str]]):
    pass


@dataclass
class ChoiceOption(Option[str]):
    choices: list[str]



"""
Abstract class for sources.

Also defines the dictionary of available sources. Each source should add itself
to this dictionary in its module.
"""

from __future__ import annotations

import asyncio
import os.path
import shlex
from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from itertools import zip_longest
from traceback import print_exc
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Type
from abc import ABC, abstractmethod


from ..log import logger
from ..entry import Entry
from ..result import Result
from ..config import BoolOption, ConfigOption


class EntryNotValid(Exception):
    """Raised when an entry is not valid for a source."""


@dataclass
class DLFilesEntry:
    """This represents a song in the context of a source.

    :param ready: This event triggers as soon, as all files for the song are
        downloaded/buffered.
    :type ready: asyncio.Event
    :param video: The location of the video part of the song.
    :type video: str
    :param audio: The location of the audio part of the song, if it is not
        incuded in the video file. (Default is ``None``)
    :type audio: Optional[str]
    :param buffering: True if parts are buffering, False otherwise (Default is
        ``False``)
    :type buffering: bool
    :param complete: True if download was completed, False otherwise (Default
        is ``False``)
    :type complete: bool
    :param skip: True if the next Entry for this file should be skipped
        (Default is ``False``)
    :param buffer_task: Reference to the task, that downloads the files.
    :type buffer_task: Optional[asyncio.Task[Tuple[str, Optional[str]]]]
    """

    # pylint: disable=too-many-instance-attributes

    ready: asyncio.Event = field(default_factory=asyncio.Event)
    video: str = ""
    audio: Optional[str] = None
    buffering: bool = False
    complete: bool = False
    skip: bool = False
    buffer_task: Optional[asyncio.Task[Tuple[str, Optional[str]]]] = None


class Source(ABC):
    """Parentclass for all sources.

    A new source should subclass this, and at least implement
    :py:func:`Source.do_buffer`, :py:func:`Song.get_entry` and
    :py:func:`Source.get_file_list`, and set the ``source_name``
    attribute.

    Source specific tasks will be forwarded to the respective source, like:
        - Buffering the audio/video
        - Searching for a query
        - Getting an entry from an identifier
        - Handling the skipping of currently played song

    Some methods of a source will be called by the server and some will be
    called by the playback client.

    Specific server methods:
    ``get_entry``, ``search``, ``add_to_config``

    Specific client methods:
    ``buffer``, ``do_buffer``, ``skip_current``, ``ensure_playable``,
    ``get_missing_metadata``, ``get_config``

    Each source has a reference to all files, that are currently queued to
    download via the :py:attr:`Source.downloaded_files` attribute and a
    reference to a ``mpv`` process playing songs for that specific source

    :attributes: - ``downloaded_files``, a dictionary mapping
                   :py:attr:`Entry.ident` to :py:class:`DLFilesEntry`.
                 - ``player``, the reference to the ``mpv`` process, if it has
                   started
                 - ``extra_mpv_options``, dictionary of arguments added to the mpv
                   instance, can be overwritten by a subclass
                 - ``source_name``, the string used to identify the source
    """

    source_name: str = ""
    config_schema: dict[str, ConfigOption[Any]] = {
        "enabled": ConfigOption(BoolOption(), "Enable this source", False)
    }

    def __init__(self, config: dict[str, Any]):
        """
        Create and initialize a new source.

        You should never try to instantiate the Source class directly, rather
        you should instantiate a subclass.

        :param config: Specific configuration for a source. See the respective
          source for documentation.
        :type config: dict[str, Any]
        """
        self.config: dict[str, Any] = config
        self.downloaded_files: defaultdict[str, DLFilesEntry] = defaultdict(DLFilesEntry)
        self._masterlock: asyncio.Lock = asyncio.Lock()
        self._index: list[str] = config["index"] if "index" in config else []
        self.extra_mpv_options: dict[str, str] = {}
        self._skip_next = False
        self.build_index = False
        self.apply_config(config)

    def is_valid(self, entry: Entry) -> bool:
        """
        Check if the entry is valid.

        Each source can implement this method to check if the entry is valid.

        :param entry: The entry to check
        :type entry: Entry
        :returns: True if the entry is valid, False otherwise.
        :rtype: bool
        """
        return True

    async def get_entry(
        self,
        performer: str,
        ident: str,
        /,
        artist: Optional[str] = None,
        title: Optional[str] = None,
    ) -> Optional[Entry]:
        """
        Create an :py:class:`syng.entry.Entry` from a given identifier.

        By default, this confirmes, that the ident is a valid entry (i.e. part
        of the indexed list), and builds an Entry by parsing the file name.

        Since the server does not have access to the actual file, only to the
        file name, ``duration`` can not be set. It will be approximated with
        180 seconds. When added to the queue, the server will ask the client
        for additional metadata, like this.

        :param performer: The performer of the song
        :type performer: str
        :param ident: Unique identifier of the song.
        :type ident: str
        :returns: New entry for the identifier, or None, if the ident is
            invalid.
        :rtype: Optional[Entry]
        :raises EntryNotValid: If the entry is not valid for the source.
        """

        res: Result = Result.from_filename(ident, self.source_name)
        entry = Entry(
            ident=ident,
            source=self.source_name,
            duration=180,
            album=res.album if res.album else "Unknown",
            title=res.title if res.title else title if title else "Unknown",
            artist=res.artist if res.artist else artist if artist else "Unknown",
            performer=performer,
            incomplete_data=True,
        )
        if not self.is_valid(entry):
            raise EntryNotValid(f"Entry {entry} is not valid for source {self.source_name}")
        return entry

    async def search(self, query: str) -> list[Result]:
        """
        Search the songs from the source for a query.

        By default, this searches in the internal index.

        :param query: The query to search for
        :type query: str
        :returns: A list of Results containing the query.
        :rtype: list[Result]
        """
        filtered: list[str] = self.filter_data_by_query(query, self._index)
        results: list[Result] = []
        for filename in filtered:
            results.append(Result.from_filename(filename, self.source_name))
        return results

    @abstractmethod
    async def do_buffer(self, entry: Entry, pos: int) -> Tuple[str, Optional[str]]:
        """
        Source specific part of buffering.

        This should asynchronous download all required files to play the entry,
        and return the location of the video and audio file. If the audio is
        included in the video file, the location for the audio file should be
        `None`.

        Abstract, needs to be implemented by subclass.

        :param entry: The entry to buffer
        :type entry: Entry
        :param pos: The position in the queue, the entry is at.
        :type pos: int
        :returns: A Tuple of the locations for the video and the audio file.
        :rtype: Tuple[str, Optional[str]]
        """

    async def buffer(self, entry: Entry, pos: int) -> None:
        """
        Buffer all necessary files for the entry.

        This calls the specific :py:func:`Source.do_buffer` method. It
        ensures, that the correct events will be triggered, when the buffer
        function ends. Also ensures, that no entry will be buffered multiple
        times.

        If this is called multiple times for the same song (even if they come
        from different entries) This will immediately return.

        :param entry: The entry to buffer
        :type entry: Entry
        :param pos: The position in the queue, the entry is at.
        :type pos: int
        :rtype: None
        """
        async with self._masterlock:
            if self.downloaded_files[entry.ident].buffering:
                return
            self.downloaded_files[entry.ident].buffering = True

        try:
            buffer_task = asyncio.create_task(self.do_buffer(entry, pos))
            self.downloaded_files[entry.ident].buffer_task = buffer_task
            video, audio = await buffer_task

            self.downloaded_files[entry.ident].video = video
            self.downloaded_files[entry.ident].audio = audio
            self.downloaded_files[entry.ident].complete = True
        except ValueError as exc:
            raise exc
        except Exception:  # pylint: disable=broad-except
            print_exc()
            raise ValueError("Buffering failed for %s" % entry)

        self.downloaded_files[entry.ident].ready.set()

    async def skip_current(self, entry: Entry) -> None:
        """
        Skips first song in the queue.

        If it is played, the player is killed, if it is still buffered, the
        buffering is aborted. Then a flag is set to keep the player from
        playing it.

        :param entry: A reference to the first entry of the queue
        :type entry: Entry
        :rtype: None
        """
        async with self._masterlock:
            self._skip_next = True
            self.downloaded_files[entry.ident].buffering = False
            buffer_task = self.downloaded_files[entry.ident].buffer_task
            if buffer_task is not None:
                buffer_task.cancel()
            self.downloaded_files[entry.ident].ready.set()

    async def ensure_playable(self, entry: Entry) -> tuple[str, Optional[str]]:
        """
        Guaranties that the given entry can be played.

        First start buffering, then wait for the buffering to end.

        :param entry: The entry to ensure playback for.
        :type entry: Entry
        :rtype: None
        """
        await self.buffer(entry, 0)
        dlfilesentry = self.downloaded_files[entry.ident]
        await dlfilesentry.ready.wait()
        return dlfilesentry.video, dlfilesentry.audio

    async def get_missing_metadata(self, _entry: Entry) -> dict[str, Any]:
        """
        Read and report missing metadata.

        If the source sended a list of filenames to the server, the server can
        search these filenames, but has no way to read e.g. the duration. This
        method will be called to return the missing metadata.

        By default this just returns an empty dict.

        :param _entry: The entry to get the metadata for
        :type _entry: Entry
        :returns: A dictionary with the missing metadata.
        :rtype dict[str, Any]
        """
        return {}

    def filter_data_by_query(self, query: str, data: list[str]) -> list[str]:
        """
        Filter the ``data``-list by the ``query``.

        :param query: The query to filter
        :type query: str
        :param data: The list to filter
        :type data: list[str]
        :return: All entries in the list containing the query.
        :rtype: list[str]
        """

        def contains_all_words(words: list[str], element: str) -> bool:
            for word in words:
                if word.lower() not in os.path.basename(element).lower():
                    return False
            return True

        splitquery = shlex.split(query)
        return [element for element in data if contains_all_words(splitquery, element)]

    async def get_file_list(self) -> list[str]:
        """
        Gather a list of all files belonging to the source.

        This list will be send to the server. When the server searches, this
        list will be searched.

        :return: List of filenames belonging to the source
        :rtype: list[str]
        """
        return []

    async def update_file_list(self) -> Optional[list[str]]:
        """
        Update the internal list of files.

        This is called after the client sends its initial file list to the
        server to update the list of files since the last time an index file
        was written.

        It should return None, if the list is already up to date.
        Otherwise it should return the new list of files.


        :rtype: Optional[list[str]]
        """
        return None

    async def update_config(self) -> Optional[dict[str, Any] | list[dict[str, Any]]]:
        """
        Update the config of the source.

        This is called after the client sends its initial config to the server to
        update the config. E.g. to update the list of files, that should be send to
        the server.

        It returns None, if the config is already up to date.
        Otherwise returns the new config.

        :rtype: Optional[dict[str, Any] | list[dict[str, Any]]
        """

        if not self.build_index:
            return None
        logger.warning(f"{self.source_name}: updating index")
        new_index = await self.update_file_list()
        logger.warning(f"{self.source_name}: done")
        if new_index is not None:
            self._index = new_index
            return await self.get_config()
        return None

    async def get_config(self) -> dict[str, Any] | list[dict[str, Any]]:
        """
        Return the part of the config, that should be send to the server.

        Can be either a dictionary or a list of dictionaries. If it is a
        dictionary, a single message will be send. If it is a list, one message
        will be send for each entry in the list.

        By default this is the list of files handled by the source, split into
        chunks of 1000 filenames. This list is cached internally, so it does
        not need to be rebuild, when the client reconnects.

        But this can be any other values, as long as the respective source can
        handle that data.

        :return: The part of the config, that should be sended to the server.
        :rtype: dict[str, Any] | list[dict[str, Any]]
        """
        packages = []
        if self.build_index:
            if not self._index:
                self._index = []
                logger.warning(f"{self.source_name}: generating index")
                self._index = await self.get_file_list()
                logger.warning(f"{self.source_name}: done")
            chunked = zip_longest(*[iter(self._index)] * 1000, fillvalue="")
            packages = [{"index": list(filter(lambda x: x != "", chunk))} for chunk in chunked]
        first_package = {
            key: value
            for key, value in self.config.items()
            if self.config_schema[key].send_to_server
        }
        if not packages:
            packages = [first_package]
        else:
            packages[0] |= first_package
        if len(packages) == 1:
            return first_package
        return packages

    def add_to_config(self, config: dict[str, Any], running_number: int) -> None:
        """
        Add the config to the own config.

        This is called on the server, if :py:func:`Source.get_config` returns a
        list.

        In the default configuration, this just adds the index key of the
        config to the index attribute of the source

        If the running_number is 0, the index will be reset.

        :param config: The part of the config to add.
        :type config: dict[str, Any]
        :param running_number: The running number of the config
        :type running_number: int
        :rtype: None
        """
        if running_number == 0:
            self._index = []
        self._index += config["index"]

    @abstractmethod
    def apply_config(self, config: dict[str, Any]) -> None:
        """
        Apply the a config to the source.

        This should be implemented by each source individually.

        :param config: The part of the config to apply.
        :type config: dict[str, Any]
        :rtype: None
        """
        pass


available_sources: dict[str, Type[Source]] = {}



"""Module for the entry of the queue."""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
import re
from typing import Any
from typing import Optional
from uuid import UUID
from uuid import uuid4


@dataclass
class Entry:
    """This represents a song in the queue.

    :param ident: An identifier, that uniquely identifies the song in its
        source.
    :type ident: str
    :param source: The name of the source, this will be played from.
    :type source: str
    :param duration: The duration of the song in seconds.
    :type duration: int
    :param title: The title of the song.
    :type title: Optional[str]
    :param artist: The name of the original artist.
    :type artist: Optional[str]
    :param album: The name of the album or compilation, this particular
        version is from.
    :type album: str
    :param performer: The person, that will sing this song.
    :type performer: str
    :param failed: A flag, that indecates, that something went wrong. E.g.
        buffering was canceled, the file could not be read from disc etc.
        The exact meaning can differ from source to source. Default is false.
    :type failed: bool
    :param skip: A flag indicating, that this song is marked for skipping.
    :type skip: bool
    :param uuid: The UUID, that identifies this exact entry in the queue.
        Will be automatically assigned on creation.
    :type uuid: UUID
    :param uid: ID of the user that added this song to the queue.
    :type uid: Optional[str]
    :param started_at: The timestamp this entry began playing. ``None``, if it
        is yet to be played.
    :type started_at: Optional[float]
    """

    # pylint: disable=too-many-instance-attributes

    ident: str
    source: str
    duration: int
    title: Optional[str]
    artist: Optional[str]
    album: str
    performer: str
    failed: bool = False
    skip: bool = False
    uuid: UUID = field(default_factory=uuid4)
    uid: Optional[str] = None
    started_at: Optional[float] = None
    incomplete_data: bool = False

    def update(self, **kwargs: Any) -> None:
        """
        Update the attributes with given substitutions.

        :param \\*\\*kwargs: Keywords taken from the list of attributes.
        :type \\*\\*kwargs: Any
        :rtype: None
        """
        self.__dict__.update(kwargs)

    def shares_performer(self, other_performer: str) -> bool:
        """
        Check if this entry shares a performer with another entry.

        :param other_performer: The performer to check against.
        :type other_performer: str
        :return: True if the performers intersect, False otherwise.
        :rtype: bool
        """

        def normalize(performers: str) -> set[str]:
            return set(
                filter(
                    lambda x: len(x) > 0 and x not in ["der", "die", "das", "alle", "und"],
                    re.sub(
                        r"[^a-zA-Z0-9\s]",
                        "",
                        re.sub(
                            r"\s",
                            " ",
                            performers.lower().replace(".", " ").replace(",", " "),
                        ),
                    ).split(" "),
                )
            )

        e1_split_names = normalize(self.performer)
        e2_split_names = normalize(other_performer)

        return len(e1_split_names.intersection(e2_split_names)) > 0



"""
Module for the Server.

The server listens for incoming connections from playback clients and web
clients via the socket.io protocol.

It manages multiple independent rooms, each with its own queue and configuration.
If configured, the server can be in private mode, where only playback clients with
a valid registration key can connect. It can also be in restricted mode, where only
search is forwarded to the playback client, unless the client has a valid registration
key.
"""

from __future__ import annotations

import asyncio
import datetime
import hashlib
import logging
import os
import random
import string
import uuid
from json.decoder import JSONDecodeError
from argparse import Namespace
from dataclasses import dataclass
from dataclasses import field
from typing import Any, Callable, Literal, AsyncGenerator, Optional, cast

import socketio
from socketio.exceptions import ConnectionRefusedError
from aiohttp import web

try:
    from profanity_check import predict
except ImportError:
    # If the profanity_check package is not installed, use a dummy function
    def predict(strings: list[str]) -> list[Literal[0, 1]]:
        return [0]


from syng.sources.source import EntryNotValid

from .result import Result

from . import jsonencoder
from . import SYNG_VERSION, SYNG_PROTOCOL_VERSION
from .log import logger
from .entry import Entry
from .queue import Queue
from .sources import available_sources
from .sources import Source

DEFAULT_CONFIG = {
    "preview_duration": 3,
    "waiting_room_policy": None,
    "last_song": None,
}


def with_state(handler: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator that forwards the state of a room to a handler.

    :param forward_room: Either the handler to decorate or a boolean
        defining if the room should be forwarded.
    :type forward_room: bool | Callable[..., Any]
    :return: The decorated handler or a function that decorates a handler
    :rtype: Callable[..., Any] | Callable[[Callable[..., Any]], Callable[..., Any]]
    """

    async def wrapper(self: Server, sid: str, *args: Any, **kwargs: Any) -> Any:
        async with self.sio.session(sid) as session:
            room = session["room"]
        state = self.clients[room]
        return await handler(self, state, sid, *args, **kwargs)

    return wrapper


def admin(handler: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator, that requires the client to be an admin.

    If the client is not an admin, the handler is not called.

    :param handler: The handler to decorate
    :type handler: Callable[..., Any]
    :return: The decorated handler
    :rtype: Callable[..., Any]
    """

    async def wrapper(self: Server, sid: str, *args: Any, **kwargs: Any) -> Any:
        async with self.sio.session(sid) as session:
            room = session["room"]
            if room not in self.clients or not await self.is_admin(self.clients[room], sid):
                await self.sio.emit("err", {"type": "NO_ADMIN"}, sid)
                return
        return await handler(self, sid, *args, **kwargs)

    return wrapper


def playback(handler: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator, that requires the client to be a playback client.

    If the client is not a playback client, the handler is not called.

    :param handler: The handler to decorate
    :type handler: Callable[..., Any]
    :return: The decorated handler
    :rtype: Callable[..., Any]
    """

    async def wrapper(self: Server, sid: str, *args: Any, **kwargs: Any) -> Any:
        async with self.sio.session(sid) as session:
            room = session["room"]
        state = self.clients[room]
        if sid != state.sid:
            return
        return await handler(self, sid, *args, **kwargs)

    return wrapper


@dataclass
class Client:
    """This stores the configuration of a specific playback client.

    In case a new playback client connects to a room, these values can be
    overwritten.

    :param sources: A dictionary mapping the name of the used sources to their
        instances.
    :type sources: Source
    :param sources_prio: A list defining the order of the search results.
    :type sources_prio: list[str]
    :param config: Various configuration options for the client:
        * `preview_duration` (`Optional[int]`): The duration in seconds the
            playback client shows a preview for the next song. This is accounted for
            in the calculation of the ETA for songs later in the queue.
        * `last_song` (`Optional[float]`): A timestamp, defining the end of the queue.
        * `waiting_room_policy` (Optional[str]): One of:
            - `forced`, if a performer is already in the queue, they are put in the
                       waiting room.
            - `optional`, if a performer is already in the queue, they have the option
                          to be put in the waiting room.
            - `None`, performers are always added to the queue.
    :type config: dict[str, Any]:
    """

    sources: dict[str, Source]
    sources_prio: list[str]
    config: dict[str, Any]


@dataclass
class State:
    """This defines the state of one session/room.

    :param secret: The secret for the room. Used to log in as an admin on the
        webclient or reconnect a playbackclient
    :type secret: str
    :param queue: A queue of :py:class:`syng.entry.Entry` objects. New songs
        are appended to this, and if a playback client requests a song, it is
        taken from the top.
    :type queue: Queue
    :param waiting_room: Contains the Entries, that are hold back, until a
        specific song is finished.
    :type waiting_room: list[Entry]
    :param recent: A list of already played songs in order.
    :type recent: list[Entry]
    :param sid: The socket.io session id of the (unique) playback client. Once
        a new playback client connects to a room (with the correct secret),
        this will be swapped with the new sid.
    :type sid: str
    :param client: The config for the playback client
    :type client: Client
    :param last_seen: Timestamp of the last connected client. Used to determine
        if a room is still in use.
    :type last_seen: datetime
    """

    queue: Queue
    waiting_room: list[Entry]
    recent: list[Entry]
    sid: str
    client: Client
    last_seen: datetime.datetime = field(init=False, default_factory=datetime.datetime.now)


class Server:
    def __init__(self) -> None:
        self.sio = socketio.AsyncServer(
            cors_allowed_origins="*", logger=True, engineio_logger=False, json=jsonencoder
        )
        self.app = web.Application()
        self.runner = web.AppRunner(self.app)
        self.admin_app = web.Application()
        self.admin_runner = web.AppRunner(self.admin_app)
        self.clients: dict[str, State] = {}
        self.sio.attach(self.app)
        self.register_handlers()

    def register_handlers(self) -> None:
        self.sio.on("get-state", self.handle_get_state)
        self.sio.on("waiting-room-append", self.handle_waiting_room_append)
        self.sio.on("show-config", self.handle_show_config)
        self.sio.on("update-config", self.handle_update_config)
        self.sio.on("append", self.handle_append)
        self.sio.on("append-anyway", self.handle_append_anyway)
        self.sio.on("meta-info", self.handle_meta_info)
        self.sio.on("get-first", self.handle_get_first)
        self.sio.on("waiting-room-to-queue", self.handle_waiting_room_to_queue)
        self.sio.on("queue-to-waiting-room", self.handle_queue_to_waiting_room)
        self.sio.on("pop-then-get-next", self.handle_pop_then_get_next)
        self.sio.on("register-client", self.handle_register_client)
        self.sio.on("sources", self.handle_sources)
        self.sio.on("config-chunk", self.handle_config_chunk)
        self.sio.on("config", self.handle_config)
        self.sio.on("register-web", self.handle_register_web)
        self.sio.on("register-admin", self.handle_register_admin)
        self.sio.on("remove-room", self.handle_remove_room)
        self.sio.on("skip-current", self.handle_skip_current)
        self.sio.on("move-to", self.handle_move_to)
        self.sio.on("move-up", self.handle_move_up)
        self.sio.on("skip", self.handle_skip)
        self.sio.on("disconnect", self.handle_disconnect)
        self.sio.on("connect", self.handle_connect)
        self.sio.on("search", self.handle_search)
        self.sio.on("search-results", self.handle_search_results)
        self.sio.on("import-queue", self.handle_import_queue)

    async def is_admin(self, state: State, sid: str) -> bool:
        """
        Check if a given sid is an admin in a room.

        :param room: The room to check
        :type room: str
        :param sid: The session id to check
        :type sid: str
        :return: True if the sid is an admin in the room, False otherwise
        :rtype: bool
        """
        if state.sid == sid:
            return True
        async with self.sio.session(sid) as session:
            return "admin" in session and session["admin"]

    async def root_handler(self, request: Any) -> Any:
        """
        Handle the index and favicon requests.

        If the path of the request ends with "/favicon.ico" return the favicon,
        otherwise the index.html. This way the javascript can read the room code
        from the url.

        :param request Any: Webrequest from aiohttp
        :return: Either the favicon or the index.html
        :rtype web.FileResponse:
        """
        if request.path.endswith("/favicon.ico"):
            return web.FileResponse(os.path.join(self.app["root_folder"], "favicon.ico"))
        return web.FileResponse(os.path.join(self.app["root_folder"], "index.html"))

    def get_number_connections(self) -> int:
        """
        Get the number of connections to the server.

        :return: The number of connections
        :rtype: int
        """
        num = 0
        for namespace in self.sio.manager.get_namespaces():
            for room in self.sio.manager.rooms[namespace]:
                for participant in self.sio.manager.get_participants(namespace, room):
                    num += 1
        return num

    def get_connections(self) -> dict[str, dict[str, list[tuple[str, str]]]]:
        """
        Get all connections to the server.

        :return: A dictionary mapping namespaces to rooms and participants.
        :rtype: dict[str, dict[str, list[tuple[str, str]]]]
        """
        connections: dict[str, dict[str, list[tuple[str, str]]]] = {}
        for namespace in self.sio.manager.get_namespaces():
            connections[namespace] = {}
            for room in self.sio.manager.rooms[namespace]:
                connections[namespace][room] = []
                for participant in self.sio.manager.get_participants(namespace, room):
                    connections[namespace][room].append(participant)
        return connections

    async def get_clients(self, room: str) -> list[dict[str, Any]]:
        """
        Get the number of clients in a room.

        :param room: The room to get the number of clients for
        :type room: str
        :return: The number of clients in the room
        :rtype: int
        """
        clients = []
        for sid, client_id in self.sio.manager.get_participants("/", room):
            client: dict[str, Any] = {}
            client["sid"] = sid
            if sid == self.clients[room].sid:
                client["type"] = "playback"
            else:
                client["type"] = "web"
            client["admin"] = await self.is_admin(self.clients[room], sid)
            clients.append(client)
        return clients

    async def admin_handler(self, request: Any) -> Any:
        """
        Handle the admin request.
        """

        rooms = [
            {
                "room": room,
                "sid": state.sid,
                "last_seen": state.last_seen.isoformat(),
                "queue": state.queue.to_list(),
                "waiting_room": state.waiting_room,
                "clients": await self.get_clients(room),
            }
            for room, state in self.clients.items()
        ]
        info_dict = {
            "version": SYNG_VERSION,
            "protocol_version": SYNG_PROTOCOL_VERSION,
            "num_connections": self.get_number_connections(),
            "connections": self.get_connections(),
            "rooms": rooms,
        }
        return web.json_response(info_dict, dumps=jsonencoder.dumps)

    async def broadcast_state(
        self, state: State, /, sid: Optional[str] = None, room: Optional[str] = None
    ) -> None:
        if room is None:
            sid = state.sid if sid is None else sid
            async with self.sio.session(sid) as session:
                room = cast(str, session["room"])
        await self.send_state(state, room)

    async def log_to_playback(self, state: State, msg: str, level: str = "info") -> None:
        """
        Log a message to the playback client.

        This is used to inform the playback client of errors or other messages.

        :param state: The state of the room
        :type state: State
        :param msg: The message to send
        :type msg: str
        :rtype: None
        """
        await self.sio.emit("msg", {"msg": msg, "type": level}, room=state.sid)

    async def send_state(self, state: State, sid: str) -> None:
        """
        Send the current state (queue and recent-list) to sid.

        This sends a "state" message. This can be received either by the playback
        client, a web client or the whole room.

        If it is send to a playback client, it will be handled by the
        :py:func:`syng.client.handle_state` function.

        :param state: The state to send
        :type state: State
        :param sid: The recepient of the "state" message
        :type sid: str:
        :rtype: None
        """

        safe_config = {k: v for k, v in state.client.config.items() if k not in ["secret", "key"]}

        await self.sio.emit(
            "state",
            {
                "queue": state.queue,
                "recent": state.recent,
                "waiting_room": state.waiting_room,
                "config": safe_config,
            },
            room=sid,
        )

    @with_state
    async def handle_get_state(self, state: State, sid: str) -> None:
        """
        Handle the "get-state" message.

        Sends the current state to whoever requests it. This failes if the sender
        is not part of any room.

        :param sid: The initial sender, and therefore recepient of the "state"
            message
        :type sid: str
        :rtype: None
        """
        await self.send_state(state, sid)

    @with_state
    async def handle_waiting_room_append(
        self, state: State, sid: str, data: dict[str, Any]
    ) -> Optional[str]:
        """
        Append a song to the waiting room.

        This should be called from a web client. Appends the entry, that is encoded
        within the data to the waiting room of the room the client is currently
        connected to.

        :param sid: The session id of the client sending this request
        :type sid: str
        :param data: A dictionary encoding the entry, that should be added to the
            waiting room.
        :type data: dict[str, Any]
        :return: The uuid of the added entry or None if the entry could not be added
        :rtype: Optional[str]
        """
        source_obj = state.client.sources[data["source"]]
        try:
            entry = await source_obj.get_entry(
                data["performer"], data["ident"], artist=data["artist"], title=data["title"]
            )
            if entry is None:
                await self.sio.emit(
                    "msg",
                    {
                        "msg": f"Unable to add to the waiting room: {data['ident']}. Maybe try again?"
                    },
                    room=sid,
                )
                return None
        except EntryNotValid as e:
            await self.sio.emit(
                "msg",
                {"msg": f"Unable to add to the waiting room: {data['ident']}. {e}"},
                room=sid,
            )
            return None

        if "uid" not in data or (
            (data["uid"] is not None and len(list(state.queue.find_by_uid(data["uid"]))) == 0)
            or (data["uid"] is None and state.queue.find_by_name(data["performer"]) is None)
        ):
            await self.append_to_queue(state, entry, sid)
            return None

        entry.uid = data["uid"]

        state.waiting_room.append(entry)
        await self.broadcast_state(state, sid=sid)
        await self.sio.emit(
            "get-meta-info",
            entry,
            room=state.sid,
        )
        return str(entry.uuid)

    async def append_to_queue(
        self, state: State, entry: Entry, report_to: Optional[str] = None
    ) -> None:
        """
        Append a song to the queue for a given session.

        Checks, if the computed start time is before the configured end time of the
        event, and reports an error, if the end time is exceeded.

        :param room: The room with the queue.
        :type room: str
        :param entry: The entry that contains the song.
        :type entry: Entry
        :param report_to: If an error occurs, who to report to.
        :type report_to: Optional[str]
        :rtype: None
        """
        first_song = state.queue.try_peek()
        if first_song is None or first_song.started_at is None:
            start_time = datetime.datetime.now().timestamp()
        else:
            start_time = first_song.started_at

        start_time = state.queue.fold(
            lambda item, time: time + item.duration + state.client.config["preview_duration"] + 1,
            start_time,
        )

        if (report_to is None or not await self.is_admin(state, report_to)) and state.client.config[
            "last_song"
        ]:
            if state.client.config["last_song"] < start_time:
                if report_to is not None:
                    await self.sio.emit(
                        "err",
                        {
                            "type": "QUEUE_FULL",
                            "end_time": state.client.config["last_song"],
                        },
                        room=report_to,
                    )
                return

        state.queue.append(entry)
        await self.broadcast_state(state, sid=report_to)

        await self.sio.emit(
            "get-meta-info",
            entry,
            room=state.sid,
        )

    @admin
    @with_state
    async def handle_show_config(self, state: State, sid: str) -> None:
        """
        Sends public config to webclient.

        This will only be send if the client is on an admin connection.

        :param sid: The session id of the client sending this request
        :type sid: str
        :rtype: None
        """
        await self.sio.emit(
            "config",
            state.client.config,
            sid,
        )

    @admin
    @with_state
    async def handle_update_config(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Forwards an updated config from an authorized webclient to the playback client.

        This is currently untrested and should be used with caution.

        :param sid: The session id of the client sending this request
        :type sid: str
        :param data: A dictionary encoding the new configuration
        :type data: dict[str, Any]
        :rtype: None
        """
        try:
            config = jsonencoder.loads(data["config"])
            await self.sio.emit(
                "update_config",
                DEFAULT_CONFIG | config,
                state.sid,
            )
            state.client.config = DEFAULT_CONFIG | config
        except JSONDecodeError:
            await self.sio.emit("err", {"type": "JSON_MALFORMED"}, room=sid)

    @with_state
    async def handle_append(self, state: State, sid: str, data: dict[str, Any]) -> Optional[str]:
        """
        Handle the "append" message.

        This should be called from a web client. Appends the entry, that is encoded
        within the data to the room the client is currently connected to. An entry
        constructed this way, will be given a UUID, to differentiate it from other
        entries for the same song. Additionally an id of the web client is saved
        for that entry.

        If the room is configured to no longer accept songs past a certain time
        (via the :py:attr:`Config.last_song` attribute), it is checked, if the
        start time of the song would exceed this time. If this is the case, the
        request is denied and a "msg" message is send to the client, detailing
        this.

        If a waitingroom is forced or optional, it is checked, if one of the performers is
        already in queue. In that case, a "ask_for_waitingroom" message is send to the
        client.

        Otherwise the song is added to the queue. And all connected clients (web
        and playback client) are informed of the new state with a "state" message.

        Since some properties of a song can only be accessed on the playback
        client, a "get-meta-info" message is send to the playback client. This is
        handled there with the :py:func:`syng.client.handle_get_meta_info`
        function.

        :param sid: The session id of the client sending this request
        :type sid: str
        :param data: A dictionary encoding the entry, that should be added to the
            queue.
        :type data: dict[str, Any]
        :return: The uuid of the added entry or None if the entry could not be added
        :rtype: Optional[str]
        """
        if len(data["performer"]) > 50:
            await self.sio.emit("err", {"type": "NAME_LENGTH", "name": data["performer"]}, room=sid)
            return None

        if predict([data["performer"]]) == [1]:
            await self.sio.emit("err", {"type": "PROFANITY", "name": data["performer"]}, room=sid)
            return None

        if state.client.config["waiting_room_policy"] and (
            state.client.config["waiting_room_policy"].lower() == "forced"
            or state.client.config["waiting_room_policy"].lower() == "optional"
        ):
            old_entry = state.queue.find_by_name(data["performer"])
            if old_entry is not None:
                await self.sio.emit(
                    "ask_for_waitingroom",
                    {
                        "current_entry": {
                            "source": data["source"],
                            "performer": data["performer"],
                            "ident": data["ident"],
                            "artist": data.get("artist", None),
                            "title": data.get("title", None),
                        },
                        "old_entry": {
                            "artist": old_entry.artist,
                            "title": old_entry.title,
                            "performer": old_entry.performer,
                        },
                    },
                    room=sid,
                )
                return None

        source_obj = state.client.sources[data["source"]]

        try:
            entry = await source_obj.get_entry(
                data["performer"],
                data["ident"],
                artist=data.get("artist", None),
                title=data.get("title", None),
            )
            if entry is None:
                await self.sio.emit(
                    "msg",
                    {"msg": f"Unable to append {data['ident']}. Maybe try again?"},
                    room=sid,
                )
                return None
        except EntryNotValid as e:
            await self.sio.emit(
                "msg",
                {"msg": f"Unable to append {data['ident']}. {e}"},
                room=sid,
            )
            return None

        logger.debug(f"Appending {entry} to queue in room {state.sid}")
        entry.uid = data["uid"] if "uid" in data else None

        await self.append_to_queue(state, entry, sid)
        return str(entry.uuid)

    @with_state
    async def handle_append_anyway(
        self, state: State, sid: str, data: dict[str, Any]
    ) -> Optional[str]:
        """
        Appends a song to the queue, even if the performer is already in queue.

        Works the same as handle_append, but without the check if the performer is already
        in queue.

        Only if the waiting_room_policy is not configured as forced.

        :return: The uuid of the added entry or None if the entry could not be added
        :rtype: Optional[str]
        """
        if len(data["performer"]) > 50:
            await self.sio.emit("err", {"type": "NAME_LENGTH", "name": data["performer"]}, room=sid)
            return None

        if predict([data["performer"]]) == [1]:
            await self.sio.emit("err", {"type": "PROFANITY", "name": data["performer"]}, room=sid)
            return None

        if state.client.config["waiting_room_policy"].lower() == "forced":
            await self.sio.emit(
                "err",
                {"type": "WAITING_ROOM_FORCED"},
                room=sid,
            )
            return None

        source_obj = state.client.sources[data["source"]]

        try:
            entry = await source_obj.get_entry(
                data["performer"], data["ident"], artist=data["artist"], title=data["title"]
            )

            if entry is None:
                await self.sio.emit(
                    "msg",
                    {"msg": f"Unable to append {data['ident']}. Maybe try again?"},
                    room=sid,
                )
                return None
        except EntryNotValid as e:
            await self.sio.emit(
                "msg",
                {"msg": f"Unable to append {data['ident']}. {e}"},
                room=sid,
            )
            return None

        entry.uid = data["uid"] if "uid" in data else None

        await self.append_to_queue(state, entry, sid)
        return str(entry.uuid)

    @playback
    @with_state
    async def handle_meta_info(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "meta-info" message.

        Updated a :py:class:syng.entry.Entry`, that is encoded in the data
        parameter, in the queue, that belongs to the room the requesting client
        belongs to, with new meta data, that is send from the playback client.

        Afterwards send the updated queue to all members of the room.

        :param sid: The session id of the client sending this request.
        :type sid: str
        :param data: A dictionary encoding the entry to update (already with the
            new metadata)
        :type data: dict[str, Any]
        :rtype: None
        """
        state.queue.update(
            data["uuid"],
            lambda item: item.update(**data["meta"], incomplete_data=False),
        )

        entry = state.queue.find_by_uuid(data["uuid"])
        if entry is not None:
            source = entry.source
            source_obj = state.client.sources[source]
            if not source_obj.is_valid(entry):
                await self.log_to_playback(
                    state, f"Entry {entry.ident} is not valid.", level="error"
                )
                await state.queue.remove(entry)
        else:
            for entry in state.waiting_room:
                if entry.uuid == data["uuid"] or str(entry.uuid) == data["uuid"]:
                    entry.update(**data["meta"], incomplete_data=False)

        await self.broadcast_state(state, sid=sid)

    @playback
    @with_state
    async def handle_get_first(self, state: State, sid: str) -> None:
        """
        Handle the "get-first" message.

        This message is send by the playback client, once it has connected. It
        should only be send for the initial song. Each subsequent song should be
        requestet with a "pop-then-get-next" message (See
        :py:func:`handle_pop_then_get_next`).

        If no songs are in the queue for this room, this function waits until one
        is available, then notes its starting time and sends it back to the
        playback client in a "play" message. This will be handled by the
        :py:func:`syng.client.handle_play` function.

        :param sid: The session id of the requesting client
        :type sid: str
        :rtype: None
        """
        current = await state.queue.peek()
        current.started_at = datetime.datetime.now().timestamp()

        await self.sio.emit("play", current, room=sid)

    @admin
    @with_state
    async def handle_queue_to_waiting_room(
        self, state: State, sid: str, data: dict[str, Any]
    ) -> None:
        """
        Handle the "queue-to-waiting" message.

        If on an admin-connection, removes a song from the queue and appends it to
        the waiting room. If the performer has only one entry in the queue, it is
        put back into the queue immediately.

        :param sid: The session id of the requesting client
        :type sid: str
        :rtype: None
        """

        entry = state.queue.find_by_uuid(data["uuid"])
        if entry is not None:
            performer_entries = list(state.queue.find_all_by_name(entry.performer))
            print(performer_entries)
            if len(performer_entries) == 1:
                return
            await state.queue.remove(entry)
            state.waiting_room.append(entry)
            await self.broadcast_state(state, sid=sid)

    @admin
    @with_state
    async def handle_waiting_room_to_queue(
        self, state: State, sid: str, data: dict[str, Any]
    ) -> None:
        """
        Handle the "waiting-room-to-queue" message.

        If on an admin-connection, removes a song from the waiting room and appends it to
        the queue.

        :param sid: The session id of the requesting client
        :type sid: str
        :rtype: None
        """
        entry = next(
            (wr_entry for wr_entry in state.waiting_room if str(wr_entry.uuid) == data["uuid"]),
            None,
        )
        if entry is not None:
            state.waiting_room.remove(entry)
            await self.append_to_queue(state, entry, sid)

    async def add_songs_from_waiting_room(self, state: State) -> None:
        """
        Add all songs from the waiting room, that should be added to the queue.

        A song should be added if none of its performers are already queued.

        This should be called every time a song leaves the queue.

        :param room: The room holding the queue.
        :type room: str
        :rtype: None
        """
        wrs_to_remove = []
        for wr_entry in state.waiting_room:
            if state.queue.find_by_name(wr_entry.performer) is None:
                await self.append_to_queue(state, wr_entry)
                wrs_to_remove.append(wr_entry)

        for wr_entry in wrs_to_remove:
            state.waiting_room.remove(wr_entry)

    async def discard_first(self, state: State) -> Entry:
        """
        Gets the first element of the queue, handling resulting triggers.

        This function is used to get the first element of the queue, and handle
        the resulting triggers. This includes adding songs from the waiting room,
        and updating the state of the room.

        :param room: The room to get the first element from.
        :type room: str
        :rtype: Entry
        """

        old_entry = await state.queue.popleft()

        await self.add_songs_from_waiting_room(state)

        state.recent.append(old_entry)
        state.last_seen = datetime.datetime.now()

        return old_entry

    @playback
    @with_state
    async def handle_pop_then_get_next(self, state: State, sid: str) -> None:
        """
        Handle the "pop-then-get-next" message.

        This function acts similar to the :py:func:`handle_get_first` function. The
        main difference is, that prior to sending a song to the playback client,
        the first element of the queue is discarded.

        Afterwards it follows the same steps as the handler for the "play" message,
        get the first element of the queue, annotate it with the current time,
        update everyones state and send the entry it to the playback client in a
        "play" message. This will be handled by the
        :py:func:`syng.client.handle_play` function.

        :param sid: The session id of the requesting playback client
        :type sid: str
        :rtype: None
        """
        await self.discard_first(state)
        await self.broadcast_state(state, sid=sid)

        current = await state.queue.peek()
        current.started_at = datetime.datetime.now().timestamp()
        await self.broadcast_state(state, sid=sid)

        await self.sio.emit("play", current, room=sid)

    def check_registration(self, key: str) -> bool:
        """
        Check if a given key is in the registration keyfile.

        This is used to authenticate a client, if the server is in private or
        restricted mode.

        :param key: The key to check
        :type key: str
        :return: True if the key is in the registration keyfile, False otherwise
        :rtype: bool
        """
        with open(self.app["registration-keyfile"], encoding="utf8") as f:
            raw_keys = f.readlines()
            keys = [key[:64] for key in raw_keys]

            return key in keys

    async def check_client_version(self, client_version: tuple[int, int, int], sid: str) -> bool:
        """
        Check if a given version is compatible with the server.

        :param version: The version to check
        :type version: tuple[int, int, int]
        :return: True if the version is compatible, False otherwise
        :rtype: bool
        """
        if client_version < SYNG_PROTOCOL_VERSION:
            await self.sio.emit(
                "msg",
                {"type": "error", "msg": "Client is incompatible and outdated. Please update."},
                room=sid,
            )
            await self.sio.emit(
                "client-registered",
                {"success": False, "room": None, "reason": "PROTOCOL_VERSION"},
                room=sid,
            )
            raise ConnectionRefusedError("Client is incompatible and outdated. Please update.")

        if client_version > SYNG_VERSION:
            await self.sio.emit(
                "msg",
                {"type": "error", "msg": "Server is outdated. Please update."},
                room=sid,
            )
            raise ConnectionRefusedError("Server is outdated. Please update.")

        if client_version < SYNG_VERSION:
            await self.sio.emit(
                "msg",
                {"type": "warning", "msg": "Client is compatible but outdated. Please update."},
                room=sid,
            )
        return True

    @admin
    @with_state
    async def handle_import_queue(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "import-queue" message.

        This will add entries to the queue and waiting room from the client.

        The data dictionary should have the following keys:
            - `queue`, a list of entries to import into the queue
            - `waiting_room`, a list of entries to import into the waiting room

        :param sid: The session id of the client sending this request
        :type sid: str
        :param data: A dictionary with the keys described above
        :type data: dict[str, Any]
        :rtype: None
        """

        queue_entries = [Entry(**entry) for entry in data.get("queue", [])]
        waiting_room_entries = [Entry(**entry) for entry in data.get("waiting_room", [])]
        recent_entries = [Entry(**entry) for entry in data.get("recent", [])]

        state.queue.extend(queue_entries)
        state.waiting_room.extend(waiting_room_entries)
        state.recent.extend(recent_entries)

        await self.broadcast_state(state, sid=sid)

    @admin
    @with_state
    async def handle_remove_room(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "remove-room" message.

        This will remove the room from the server, and delete all associated data.
        This is only available on an admin connection.

        :param sid: The session id of the client sending this request
        :type sid: str
        :rtype: None
        """
        async with self.sio.session(sid) as session:
            room = cast(str, session["room"])
        if room not in self.clients:
            await self.sio.emit(
                "msg",
                {"type": "error", "msg": f"Room {room} does not exist."},
                room=sid,
            )
            return

        await self.sio.emit("room-removed", {"room": room}, room=sid)
        for client, _ in self.sio.manager.get_participants("/", room):
            await self.sio.leave_room(client, room)
            await self.sio.disconnect(client)
        del self.clients[room]
        logger.info("Removed room %s", room)

    async def handle_register_client(self, sid: str, data: dict[str, Any]) -> None:
        """
        THIS IS DEPRECATED, REGISTRATION IS NOW DONE VIA THE CONNECT EVENT.

        Handle the "register-client" message.

        The data dictionary should have the following keys:
            - `room` (Optional), the requested room
            - `config`, an dictionary of initial configurations
            - `queue`, a list of initial entries for the queue. The entries are
                        encoded as a dictionary.
            - `recent`, a list of initial entries for the recent list. The entries
                        are encoded as a dictionary.
            - `secret`, the secret of the room
            - `version`, the version of the client as a triple of integers
            - `key`, a registration key given out by the server administrator

        This will register a new playback client to a specific room. If there
        already exists a playback client registered for this room, this
        playback client will be replaced if and only if, the new playback
        client has the same secret.

        If registration is restricted, abort, if the given key is not in the
        registration keyfile.

        If no room is provided, a fresh room id is generated.

        If the client provides a new room, or a new room id was generated, the
        server will create a new :py:class:`State` object and associate it with
        the room id. The state will be initialized with a queue and recent
        list, an initial config as well as no sources (yet).

        In any case, the client will be notified of the success or failure, along
        with its assigned room key via a "client-registered" message. This will be
        handled by the :py:func:`syng.client.handle_client_registered` function.

        If it was successfully registerd, the client will be added to its assigend
        or requested room.

        Afterwards all clients in the room will be send the current state.

        :param sid: The session id of the requesting playback client.
        :type sid: str
        :param data: A dictionary with the keys described above
        :type data: dict[str, Any]
        :rtype: None
        """

        # if "version" not in data:
        #     await self.sio.emit(
        #         "client-registered",
        #         {"success": False, "room": None, "reason": "NO_VERSION"},
        #         room=sid,
        #     )
        #     return
        #
        # client_version = tuple(data["version"])
        # if not await self.check_client_version(client_version, sid):
        #     return

        def gen_id(length: int = 4) -> str:
            client_id = "".join([random.choice(string.ascii_letters) for _ in range(length)])
            if client_id in self.clients:
                client_id = gen_id(length + 1)
            return client_id

        if "key" in data["config"]:
            data["config"]["key"] = hashlib.sha256(data["config"]["key"].encode()).hexdigest()

        if self.app["type"] == "private" and (
            "key" not in data["config"] or not self.check_registration(data["config"]["key"])
        ):
            await self.sio.emit(
                "client-registered",
                {
                    "success": False,
                    "room": None,
                    "reason": "PRIVATE",
                },
                room=sid,
            )
            return

        room: str = (
            data["config"]["room"]
            if "room" in data["config"] and data["config"]["room"]
            else gen_id()
        )
        async with self.sio.session(sid) as session:
            session["room"] = room

        if room in self.clients:
            old_state: State = self.clients[room]
            if data["config"]["secret"] == old_state.client.config["secret"]:
                logger.info("Got new client connection for %s", room)
                old_state.sid = sid
                old_state.client = Client(
                    sources=old_state.client.sources,
                    sources_prio=old_state.client.sources_prio,
                    config=DEFAULT_CONFIG | data["config"],
                )
                await self.sio.enter_room(sid, room)
                await self.sio.emit(
                    "client-registered",
                    {
                        "success": True,
                        "room": room,
                        "queue": self.clients[room].queue,
                        "recent": self.clients[room].recent,
                        "waiting_room": self.clients[room].waiting_room,
                    },
                    room=sid,
                )
                await self.send_state(self.clients[room], sid)
            else:
                logger.warning("Got wrong secret for %s", room)
                await self.sio.emit("client-registered", {"success": False, "room": room}, room=sid)
        else:
            logger.info("Registerd new client %s", room)
            initial_entries = [Entry(**entry) for entry in data["queue"]]
            initial_waiting_room = [Entry(**entry) for entry in data["waiting_room"]]
            initial_recent = [Entry(**entry) for entry in data["recent"]]

            self.clients[room] = State(
                queue=Queue(initial_entries),
                waiting_room=initial_waiting_room,
                recent=initial_recent,
                sid=sid,
                client=Client(
                    sources={},
                    sources_prio=[],
                    config=DEFAULT_CONFIG | data["config"],
                ),
            )

            await self.sio.enter_room(sid, room)
            await self.sio.emit(
                "client-registered",
                {
                    "success": True,
                    "room": room,
                    "queue": self.clients[room].queue,
                    "recent": self.clients[room].recent,
                    "waiting_room": self.clients[room].waiting_room,
                },
                room=sid,
            )
            # await self.send_state(self.clients[room], sid)

    @playback
    @with_state
    async def handle_sources(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "sources" message.

        Get the list of sources the client wants to use. Update internal list of
        sources, remove unused sources and query for a config for all uninitialized
        sources by sending a "request-config" message for each such source to the
        playback client. This will be handled by the
        :py:func:`syng.client.request-config` function.

        This will not yet add the sources to the configuration, rather gather what
        sources need to be configured and request their configuration. The list
        of sources will set the :py:attr:`Config.sources_prio` attribute.

        :param sid: The session id of the playback client
        :type sid: str
        :param data: A dictionary containing a "sources" key, with the list of
            sources to use.
        :type data: dict[str, Any]
        :rtype: None
        """
        unused_sources = state.client.sources.keys() - data["sources"]
        new_sources = data["sources"] - state.client.sources.keys()

        for source in unused_sources:
            del state.client.sources[source]

        state.client.sources_prio = data["sources"]

        for name in new_sources:
            await self.sio.emit("request-config", {"source": name}, room=sid)

    @playback
    @with_state
    async def handle_config_chunk(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "config-chunk" message.

        This is called, when a source wants its configuration transmitted in
        chunks, rather than a single message. If the source already exist
        (e.g. when this is not the first chunk), the config will be added
        to the source, otherwise a source will be created with the given
        configuration.

        :param sid: The session id of the playback client
        :type sid: str
        :param data: A dictionary with a "source" (str) and a
            "config" (dict[str, Any]) entry. The exact content of the config entry
            depends on the source.
        :rtype: None
        """
        if data["source"] not in state.client.sources:
            state.client.sources[data["source"]] = available_sources[data["source"]](data["config"])
        else:
            state.client.sources[data["source"]].add_to_config(data["config"], data["number"])

    @playback
    @with_state
    async def handle_config(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "config" message.

        This is called, when a source wants its configuration transmitted in
        a single message, rather than chunks. A source will be created with the
        given configuration.

        :param sid: The session id of the playback client
        :type sid: str
        :param data: A dictionary with a "source" (str) and a
            "config" (dict[str, Any]) entry. The exact content of the config entry
            depends on the source.
        :type data: dict[str, Any]
        :rtype: None
        """
        state.client.sources[data["source"]] = available_sources[data["source"]](data["config"])

    async def handle_connect(
        self, sid: str, environ: dict[str, Any], auth: None | dict[str, Any] = None
    ) -> None:
        """
        Handle the "connect" message.

        This is called, when a client connects to the server. It will register the
        client and send the initial state of the room to the client.

        :param sid: The session id of the requesting client.
        :type sid: str
        :param data: A dictionary with the keys described in
            :py:func:`handle_register_client`.
        :type data: dict[str, Any]
        :rtype: None
        """
        logger.debug("Client %s connected", sid)
        if auth is None or "type" not in auth:
            logger.warning(
                "Client %s connected without auth data, fall back to old registration", sid
            )
            return
            # raise ConnectionRefusedError("No authentication data provided. Please register first.")

        match auth["type"]:
            case "playback":
                await self.register_playback_client(sid, auth)
            case "web":
                await self.register_web_client(sid, auth)

    async def register_web_client(self, sid: str, auth: dict[str, Any]) -> None:
        if auth["room"] in self.clients:
            logger.info("Client %s registered for room %s", sid, auth["room"])
            async with self.sio.session(sid) as session:
                session["room"] = auth["room"]
                await self.sio.enter_room(sid, session["room"])
            state = self.clients[session["room"]]
            await self.send_state(state, sid)
            is_admin = False
            if "secret" in auth:
                is_admin = auth["secret"] == state.client.config["secret"]
                async with self.sio.session(sid) as session:
                    session["admin"] = is_admin
            await self.sio.emit("admin", is_admin, room=sid)
        else:
            logger.warning(
                "Client %s tried to register for non-existing room %s", sid, auth["room"]
            )
            raise ConnectionRefusedError(
                f"Room {auth['room']} does not exist. Please register first."
            )

    async def register_playback_client(self, sid: str, data: dict[str, Any]) -> None:
        """
        Register a new playback client and create a new room if necessary.

        The data dictionary should have the following keys:
            - `room` (Optional), the requested room
            - `config`, an dictionary of initial configurations
            - `queue`, a list of initial entries for the queue. The entries are
                        encoded as a dictionary.
            - `recent`, a list of initial entries for the recent list. The entries
                        are encoded as a dictionary.
            - `secret`, the secret of the room
            - `version`, the version of the client as a triple of integers
            - `key`, a registration key given out by the server administrator

        This will register a new playback client to a specific room. If there
        already exists a playback client registered for this room, this
        playback client will be replaced if and only if, the new playback
        client has the same secret.

        If registration is restricted, abort, if the given key is not in the
        registration keyfile.

        If no room is provided, a fresh room id is generated.

        If the client provides a new room, or a new room id was generated, the
        server will create a new :py:class:`State` object and associate it with
        the room id. The state will be initialized with a queue and recent
        list, an initial config as well as no sources (yet).

        In any case, the client will be notified of the success or failure, along
        with its assigned room key via a "client-registered" message. This will be
        handled by the :py:func:`syng.client.handle_client_registered` function.

        If it was successfully registerd, the client will be added to its assigend
        or requested room.

        Afterwards all clients in the room will be send the current state.

        :param sid: The session id of the requesting playback client.
        :type sid: str
        :param data: A dictionary with the keys described above
        :type data: dict[str, Any]
        :rtype: None
        """
        if "version" not in data:
            pass
            # TODO: Fallback to old registration method

            # await self.sio.emit(
            #     "client-registered",
            #     {"success": False, "room": None, "reason": "NO_VERSION"},
            #     room=sid,
            # )
            return

        client_version = tuple(data["version"])
        if not await self.check_client_version(client_version, sid):
            return

        def gen_id(length: int = 4) -> str:
            client_id = "".join([random.choice(string.ascii_letters) for _ in range(length)])
            if client_id in self.clients:
                client_id = gen_id(length + 1)
            return client_id

        if "key" in data["config"]:
            data["config"]["key"] = hashlib.sha256(data["config"]["key"].encode()).hexdigest()

        if self.app["type"] == "private" and (
            "key" not in data["config"] or not self.check_registration(data["config"]["key"])
        ):
            await self.sio.emit(
                "client-registered",
                {
                    "success": False,
                    "room": None,
                    "reason": "PRIVATE",
                },
                room=sid,
            )
            raise ConnectionRefusedError(
                "Private server, registration key not provided or invalid."
            )

        room: str = (
            data["config"]["room"]
            if "room" in data["config"] and data["config"]["room"]
            else gen_id()
        )
        async with self.sio.session(sid) as session:
            session["room"] = room

        if room in self.clients:
            old_state: State = self.clients[room]
            if data["config"]["secret"] == old_state.client.config["secret"]:
                logger.info("Got new playback client connection for %s", room)
                old_state.sid = sid
                old_state.client = Client(
                    sources=old_state.client.sources,
                    sources_prio=old_state.client.sources_prio,
                    config=DEFAULT_CONFIG | data["config"],
                )
                await self.sio.enter_room(sid, room)
                await self.send_state(self.clients[room], sid)
            else:
                logger.warning("Got wrong secret for %s", room)
                raise ConnectionRefusedError(f"Wrong secret for room {room}.")
        else:
            logger.info("Registerd new playback client %s", room)
            initial_entries = [Entry(**entry) for entry in data["queue"]]
            initial_waiting_room = [Entry(**entry) for entry in data["waiting_room"]]
            initial_recent = [Entry(**entry) for entry in data["recent"]]

            self.clients[room] = State(
                queue=Queue(initial_entries),
                waiting_room=initial_waiting_room,
                recent=initial_recent,
                sid=sid,
                client=Client(
                    sources={},
                    sources_prio=[],
                    config=DEFAULT_CONFIG | data["config"],
                ),
            )

            await self.sio.enter_room(sid, room)
            await self.send_state(self.clients[room], sid)

    async def handle_register_web(self, sid: str, data: dict[str, Any]) -> bool:
        """
        THIS IS DEPRECATED, REGISTRATION IS NOW DONE VIA THE CONNECT EVENT.

        Handle a "register-web" message.

        Adds a web client to a requested room and sends it the initial state of the
        queue and recent list.

        :param sid: The session id of the web client.
        :type sid: str
        :param data: A dictionary, containing at least a "room" entry.
        :type data: dict[str, Any]
        :return: True, if the room exist, False otherwise
        :rtype: bool
        """
        if "room" in data and data["room"] in self.clients:
            async with self.sio.session(sid) as session:
                session["room"] = data["room"]
                await self.sio.enter_room(sid, session["room"])
            state = self.clients[session["room"]]
            await self.send_state(state, sid)
            return True
        return False

    @with_state
    async def handle_register_admin(self, state: State, sid: str, data: dict[str, Any]) -> bool:
        """
        THIS IS DEPRECATED, REGISTRATION IS NOW DONE VIA THE CONNECT EVENT.

        Handle a "register-admin" message.

        If the client provides the correct secret for its room, the connection is
        upgraded to an admin connection.

        :param sid: The session id of the client, requesting admin.
        :type sid: str:
        :param data: A dictionary with at least a "secret" entry.
        :type data: dict[str, Any]
        :return: True, if the secret is correct, False otherwise
        :rtype: bool
        """
        is_admin: bool = data["secret"] == state.client.config["secret"]
        async with self.sio.session(sid) as session:
            session["admin"] = is_admin
        return is_admin

    @admin
    @with_state
    async def handle_skip_current(self, state: State, sid: str) -> None:
        """
        Handle a "skip-current" message.

        If this comes from an admin connection, forward the "skip-current" message
        to the playback client. This will be handled by the
        :py:func:`syng.client.handle_skip_current` function.

        :param sid: The session id of the client, requesting.
        :type sid: str
        :rtype: None
        """
        old_entry = await self.discard_first(state)
        await self.sio.emit("skip-current", old_entry, room=state.sid)
        await self.broadcast_state(state, sid=sid)

    @admin
    @with_state
    async def handle_move_to(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "move-to" message.

        If on an admin connection, moves the entry specified in the data to the
        position specified in the data.

        :param sid: The session id of the client requesting.
        :type sid: str
        :param data: A dictionary with at least an "uuid" and a "target" entry
        :type data: dict[str, Any]
        :rtype: None
        """
        await state.queue.move_to(data["uuid"], data["target"])
        await self.broadcast_state(state, sid=sid)

    @admin
    @with_state
    async def handle_move_up(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "move-up" message.

        If on an admin connection, moves up the entry specified in the data by one
        place in the queue.

        :param sid: The session id of the client requesting.
        :type sid: str
        :param data: A dictionary with at least an "uuid" entry
        :type data: dict[str, Any]
        :rtype: None
        """
        await state.queue.move_up(data["uuid"])
        await self.broadcast_state(state, sid=sid)

    @admin
    @with_state
    async def handle_skip(self, state: State, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "skip" message.

        If on an admin connection, removes the entry specified by data["uuid"]
        from the queue or the waiting room. Triggers the waiting room.

        :param sid: The session id of the client requesting.
        :type sid: str
        :param data: A dictionary with at least an "uuid" entry.
        :type data: dict[str, Any]
        :rtype: None
        """
        entry = state.queue.find_by_uuid(data["uuid"])
        if entry is not None:
            logger.info("Skipping %s", entry)

            await self.add_songs_from_waiting_room(state)

            await state.queue.remove(entry)

        first_entry_index = None
        for idx, wr_entry in enumerate(state.waiting_room):
            if wr_entry.uuid == data["uuid"]:
                first_entry_index = idx
                break

        if first_entry_index is not None:
            logger.info(
                "Deleting %s from waiting room",
                state.waiting_room[first_entry_index],
            )
            del state.waiting_room[first_entry_index]
        await self.broadcast_state(state, sid=sid)

    async def handle_disconnect(self, sid: str) -> None:
        """
        Handle the "disconnect" message.

        This message is send automatically, when a client disconnets.

        Remove the client from its room.

        :param sid: The session id of the client disconnecting
        :type sid: str
        :rtype: None
        """
        async with self.sio.session(sid) as session:
            room = session.get("room")
        if room is not None:
            await self.sio.leave_room(sid, room)

    @with_state
    async def handle_search(self, state: State, sid: str, data: dict[str, Any]) -> str:
        """
        Handle the "search" message.

        Forwards the dict["query"] to the :py:func:`Source.search` method, and
        execute them concurrently. The order is given by the
        :py:attr:`Config.sources_prio` attribute of the state.

        The result will be send with a "search-results" message to the (web)
        client.

        :param sid: The session id of the client requesting.
        :type sid: str
        :param data: A dictionary with at least a "query" entry.
        :type data: dict[str, str]
        :return: The search id
        :rtype: str
        """
        query = data["query"]
        search_id = uuid.uuid4()
        if (
            self.app["type"] != "restricted"
            or "key" in state.client.config
            and self.check_registration(state.client.config["key"])
        ):
            asyncio.create_task(self.search_and_emit(search_id, query, state, sid))
        else:
            await self.sio.emit(
                "search", {"query": query, "sid": sid, "search_id": search_id}, room=state.sid
            )
        return str(search_id)

    async def search_and_emit(
        self, search_id: uuid.UUID, query: str, state: State, sid: str
    ) -> None:
        """
        Search for a query on a source and emit the results.

        :param search_id: The search id
        :type search_id: uuid.UUID
        :param query: The query to search for
        :type query: str
        :param state: The state of the room
        :type state: State
        :param sid: The session id of the client
        :type sid: str
        :rtype: list[Result]
        """

        results_list = await asyncio.gather(
            *[state.client.sources[source].search(query) for source in state.client.sources_prio]
        )

        results = [
            search_result for source_result in results_list for search_result in source_result
        ]
        await self.send_search_results(sid, results, search_id)

    @playback
    async def handle_search_results(self, sid: str, data: dict[str, Any]) -> None:
        """
        Handle the "search-results" message.

        This message is send by the playback client, once it has received search
        results. The results are send to the web client.

        The data dictionary should have the following keys:
            - `sid`, the session id of the web client (str)
            - `results`, a list of search results (list[dict[str, Any]])
            - `search_id`, the search id (str) (Optional)

        :param sid: The session id of the playback client
        :type sid: str
        :param data: A dictionary with the keys described above
        :type data: dict[str, Any]
        :rtype: None
        """
        web_sid = data["sid"]
        search_id = data["search_id"] if "search_id" in data else None

        results = [Result.from_dict(result) for result in data["results"]]

        await self.send_search_results(web_sid, results, search_id)

    async def send_search_results(
        self, sid: str, results: list[Result], search_id: Optional[uuid.UUID]
    ) -> None:
        """
        Send search results to a client.

        :param sid: The session id of the client to send the results to.
        :type sid: str
        :param results: The search results to send.
        :type results: list[Result]
        :rtype: None
        """
        await self.sio.emit(
            "search-results",
            {"results": results, "search_id": search_id},
            room=sid,
        )

    async def cleanup(self) -> None:
        """
        Clean up the unused playback clients

        This runs every hour, and removes every client, that did not requested a song for four hours.

        :rtype: None
        """

        logger.info("Start Cleanup")
        to_remove: list[str] = []
        for sid, state in self.clients.items():
            logger.debug("Client %s, last seen: %s", sid, str(state.last_seen))
            if state.last_seen + datetime.timedelta(hours=4) < datetime.datetime.now():
                logger.info("No activity for 4 hours, removing %s", sid)
                to_remove.append(sid)
        for sid in to_remove:
            await self.sio.disconnect(sid)
            del self.clients[sid]

        # The internal loop counter does not use a regular timestamp, so we need to convert between
        # regular datetime and the async loop time
        now = datetime.datetime.now()

        next_run = now + datetime.timedelta(hours=1)
        offset = next_run.timestamp() - now.timestamp()
        loop_next = asyncio.get_event_loop().time() + offset

        logger.info("End cleanup, next cleanup at %s", str(next_run))
        asyncio.get_event_loop().call_at(loop_next, lambda: asyncio.create_task(self.cleanup()))

    async def background_tasks(
        self,
        iapp: web.Application,
    ) -> AsyncGenerator[None, None]:
        """
        Create all the background tasks.

        For now, this is only the cleanup task.

        :param iapp: The web application
        :type iapp: web.Application
        :rtype: AsyncGenerator[None, None]
        """

        iapp["repeated_cleanup"] = asyncio.create_task(self.cleanup())

        yield

        iapp["repeated_cleanup"].cancel()
        await iapp["repeated_cleanup"]

    async def run_apps(self, host: str, port: int, admin_port: Optional[int]) -> None:
        """
        Run the main and admin apps.

        This is used to run the main app and the admin app in parallel.

        :param host: The host to bind to
        :type host: str
        :param port: The port to bind to
        :type port: int
        :param admin_port: The port for the admin interface, or None if not used
        :type admin_port: Optional[int]
        :rtype: None
        """
        if admin_port:
            logger.info("Starting admin interface on port %d", admin_port)
            print(f"==== Admin Interface on {host}:{admin_port} ====")
            await self.admin_runner.setup()
            admin_site = web.TCPSite(self.admin_runner, host, admin_port)
            await admin_site.start()
        logger.info("Starting main server on port %d", port)
        print(f"==== Server on {host}:{port} ====")
        await self.runner.setup()
        site = web.TCPSite(self.runner, host, port)
        await site.start()

        while True:
            await asyncio.sleep(3600)

    def run(self, args: Namespace) -> None:
        """
        Run the server.

        `args` consists of the following attributes:
            - `host`, the host to bind to
            - `port`, the port to bind to
            - `root_folder`, the root folder of the web client
            - `registration_keyfile`, the file containing the registration keys
            - `private`, if the server is private
            - `restricted`, if the server is restricted
            - `admin_port`, the port for the admin interface

        :param args: The command line arguments
        :type args: Namespace
        :rtype: None
        """
        self.app["type"] = (
            "private" if args.private else "restricted" if args.restricted else "public"
        )
        if args.registration_keyfile:
            self.app["registration-keyfile"] = args.registration_keyfile

        self.app["root_folder"] = args.root_folder

        self.app.add_routes(
            [web.static("/assets/", os.path.join(self.app["root_folder"], "assets/"))]
        )

        self.app.router.add_route("*", "/", self.root_handler)
        self.app.router.add_route("*", "/{room}", self.root_handler)
        self.app.router.add_route("*", "/{room}/", self.root_handler)
        self.admin_app.router.add_route("*", "/", self.admin_handler)

        self.app.cleanup_ctx.append(self.background_tasks)
        if args.admin_password:
            self.sio.instrument(auth={"username": "admin", "password": args.admin_password})

        try:
            asyncio.run(
                self.run_apps(
                    args.host,
                    args.port,
                    args.admin_port,
                )
            )
        except KeyboardInterrupt:
            pass
        finally:
            logger.info("Shutting down server...")
            asyncio.run(self.runner.cleanup())
            asyncio.run(self.admin_runner.cleanup())
            logger.info("Server shut down.")


def run_server(args: Namespace) -> None:
    """
    Run the server.

    :param args: The command line arguments
    :type args: Namespace
    :rtype: None
    """
    loglevel = getattr(logging, args.log_level.upper(), logging.WARNING)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    logger.setLevel(loglevel)
    server = Server()
    server.run(args)



"""A async queue with synchronization."""

import asyncio
from collections import deque
from collections.abc import Callable, Iterable
from typing import Any
from typing import Optional
from uuid import UUID

from .entry import Entry


class Queue:
    """A async queue with synchronization.

    This queue keeps track of the amount of entries by using a semaphore.

    :param initial_entries: Initial list of entries to add to the queue
    :type initial_entries: list[Entry]
    """

    def __init__(self, initial_entries: list[Entry]):
        """
        Construct the queue. And initialize the internal lock and semaphore.

        :param initial_entries: Initial list of entries to add to the queue
        :type initial_entries: list[Entry]
        """
        self._queue = deque(initial_entries)

        self.num_of_entries_sem = asyncio.Semaphore(len(self._queue))
        self.readlock = asyncio.Lock()

    def extend(self, entries: Iterable[Entry]) -> None:
        """
        Extend the queue with a list of entries and increase the semaphore.

        :param entries: The entries to add
        :type entries: Iterable[Entry]
        :rtype: None
        """
        for entry in entries:
            self.append(entry)

    def append(self, entry: Entry) -> None:
        """
        Append an entry to the queue, increase the semaphore.

        :param entry: The entry to add
        :type entry: Entry
        :rtype: None
        """
        self._queue.append(entry)
        self.num_of_entries_sem.release()

    def try_peek(self) -> Optional[Entry]:
        """Return the first entry in the queue, if it exists."""
        if len(self._queue) > 0:
            return self._queue[0]
        return None

    async def peek(self) -> Entry:
        """
        Return the first entry in the queue.

        If the queue is empty, wait until the queue has at least one entry.

        :returns: First entry of the queue
        :rtype: Entry
        """
        async with self.readlock:
            await self.num_of_entries_sem.acquire()
            item = self._queue[0]
            self.num_of_entries_sem.release()
        return item

    async def popleft(self) -> Entry:
        """
        Remove the first entry in the queue and return it.

        Decreases the semaphore. If the queue is empty, wait until the queue
        has at least one entry.

        :returns: First entry of the queue
        :rtype: Entry
        """
        async with self.readlock:
            await self.num_of_entries_sem.acquire()
            item = self._queue.popleft()
        return item

    def to_list(self) -> list[Entry]:
        """
        Return all entries in a list.

        This is done, so that the entries can be converted to a JSON object,
        when sending it to the web or playback client.

        :returns: A list with all the entries.
        :rtype: list[Entry]
        """
        return list(self._queue)  # [item for item in self._queue]

    def update(self, uuid: UUID | str, updater: Callable[[Entry], None]) -> None:
        """
        Update entries in the queue, identified by their uuid.

        If an entry with that uuid is not in the queue, nothing happens.

        :param uuid: The uuid of the entry to update
        :type uuid: UUID | str
        :param updater: A function, that updates the entry
        :type updater: Callable[[Entry], None]
        :rtype: None
        """
        for item in self._queue:
            if item.uuid == uuid or str(item.uuid) == uuid:
                updater(item)

    def find_by_name(self, name: str) -> Optional[Entry]:
        """
        Find the first entry by its performer and return it.

        :param name: The name of the performer to search for.
        :type name: str
        :returns: The entry with the performer or `None` if no such entry exists
        :rtype: Optional[Entry]
        """
        for item in self._queue:
            if item.shares_performer(name):
                return item
        return None

    def find_all_by_name(self, name: str) -> Iterable[Entry]:
        """
        Find all entries by their performer and return them as an iterable.

        :param name: The name of the performer to search for.
        :type name: str
        :returns: The entries with the performer.
        :rtype: Iterable[Entry]
        """

        for item in self._queue:
            if item.shares_performer(name):
                yield item

    def find_by_uuid(self, uuid: UUID | str) -> Optional[Entry]:
        """
        Find an entry by its uuid and return it.

        :param uuid: The uuid to search for.
        :type uuid: UUID | str
        :returns: The entry with the uuid or `None` if no such entry exists
        :rtype: Optional[Entry]
        """
        for item in self._queue:
            if item.uuid == uuid or str(item.uuid) == uuid:
                return item
        return None

    def find_by_uid(self, uid: str) -> Iterable[Entry]:
        """
        Find all entries for a given user id
        """

        for item in self._queue:
            if item.uid == uid:
                yield item

    def fold(self, func: Callable[[Entry, Any], Any], start_value: Any) -> Any:
        """Call ``func`` on each entry and accumulate the result."""
        for item in self._queue:
            start_value = func(item, start_value)
        return start_value

    async def remove(self, entry: Entry) -> None:
        """
        Remove an entry, if it exists. Decrease the semaphore.

        :param entry: The entry to remove
        :type entry: Entry
        :rtype: None
        """
        async with self.readlock:
            await self.num_of_entries_sem.acquire()
            self._queue.remove(entry)

    async def move_up(self, uuid: str) -> None:
        """
        Move an :py:class:`syng.entry.Entry` with the uuid up in the queue.

        If it is called on the first two elements, nothing will happen.

        :param uuid: The uuid of the entry.
        :type uuid: str
        :rtype: None
        """
        async with self.readlock:
            uuid_idx = 0
            for idx, item in enumerate(self._queue):
                if item.uuid == uuid or str(item.uuid) == uuid:
                    uuid_idx = idx

            if uuid_idx > 1:
                tmp = self._queue[uuid_idx]
                self._queue[uuid_idx] = self._queue[uuid_idx - 1]
                self._queue[uuid_idx - 1] = tmp

    async def move_to(self, uuid: str, target: int) -> None:
        """
        Move an :py:class:`syng.entry.Entry` with the uuid to a specific position.

        :param uuid: The uuid of the entry.
        :type uuid: str
        :param target: The target position.
        :type target: int
        :rtype: None
        """

        async with self.readlock:
            uuid_idx = 0
            for idx, item in enumerate(self._queue):
                if item.uuid == uuid or str(item.uuid) == uuid:
                    uuid_idx = idx

            if uuid_idx != target:
                entry = self._queue[uuid_idx]
                self._queue.remove(entry)

                if target > uuid_idx:
                    target = target - 1
                self._queue.insert(target, entry)



"""
Main entry point for the application.

This module contains the main entry point for the application. It parses the
command line arguments and runs the appropriate function based on the arguments.

This module also checks if the client and server modules are available and
imports them if they are. If they are not available, the application will not
run the client or server functions.

Client usage: syng client [-h] [--room ROOM] [--secret SECRET] \
                    [--config-file CONFIG_FILE] [--server SERVER]
Server usage: syng server [-h] [--host HOST] [--port PORT] [--root-folder ROOT_FOLDER] \
                    [--registration-keyfile REGISTRATION_KEYFILE] [--private] [--restricted] \
                    [--admin-password PASSWORD]
GUI usage: syng gui

The config file for the client should be a yaml file in the following style::

      sources:
        SOURCE1:  
          configuration for SOURCE
        SOURCE2: 
          configuration for SOURCE
        ...
      config:
        server: ...
        room: ...
        preview_duration: ...
        secret: ...
        last_song: ...
        waiting_room_policy: ..
        key: ..
"""

from typing import TYPE_CHECKING
from argparse import ArgumentParser
import os
import multiprocessing
import traceback

import platformdirs

gui_exception = ""
try:
    from syng.gui import run_gui

    GUI_AVAILABLE = True
except ImportError:
    if TYPE_CHECKING:
        from syng.gui import run_gui
    gui_exception = traceback.format_exc()
    GUI_AVAILABLE = False

try:
    from .client import run_client

    CLIENT_AVAILABLE = True
except ImportError:
    if TYPE_CHECKING:
        from .client import run_client

    CLIENT_AVAILABLE = False

try:
    from .server import run_server

    SERVER_AVAILABLE = True
except ImportError:
    if TYPE_CHECKING:
        from .server import run_server

    SERVER_AVAILABLE = False


def main() -> None:
    """
    Main entry point for the application.

    This function parses the command line arguments and runs the appropriate
    function based on the arguments.

    :return: None
    """
    parser: ArgumentParser = ArgumentParser()
    sub_parsers = parser.add_subparsers(dest="action")

    if CLIENT_AVAILABLE:
        client_parser = sub_parsers.add_parser("client")

        client_parser.add_argument("--room", "-r")
        client_parser.add_argument("--secret", "-s")
        client_parser.add_argument(
            "--config-file",
            "-C",
            default=f"{os.path.join(platformdirs.user_config_dir('syng'), 'config.yaml')}",
        )
        # client_parser.add_argument("--key", "-k", default=None)
        client_parser.add_argument("--server", "-S")

    if GUI_AVAILABLE:
        sub_parsers.add_parser("gui")

    if SERVER_AVAILABLE:
        root_path = os.path.join(os.path.dirname(__file__), "static")
        server_parser = sub_parsers.add_parser("server")
        server_parser.add_argument("--host", "-H", default="localhost")
        server_parser.add_argument("--port", "-p", type=int, default=8080)
        server_parser.add_argument("--root-folder", "-r", default=root_path)
        server_parser.add_argument("--registration-keyfile", "-k", default=None)
        server_parser.add_argument("--private", "-P", action="store_true", default=False)
        server_parser.add_argument("--restricted", "-R", action="store_true", default=False)
        server_parser.add_argument("--admin-password", "-A", default=None)
        server_parser.add_argument("--admin-port", "-a", type=int, default=None)
        server_parser.add_argument(
            "--log-level",
            "-l",
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "FATAL"],
        )

    args = parser.parse_args()

    if args.action == "client":
        run_client(args)
    elif args.action == "server":
        run_server(args)
    elif args.action == "gui":
        if not GUI_AVAILABLE:
            print("GUI module is not available.")
            print(gui_exception)
        else:
            run_gui()
    else:
        if not GUI_AVAILABLE:
            print("GUI module is not available.")
            print(gui_exception)
        else:
            run_gui()


if __name__ == "__main__":
    if os.name == "nt":
        multiprocessing.freeze_support()
    main()



import asyncio
from enum import Enum
import locale
import sys
from typing import Any, Callable, Iterable, Optional, cast
from qrcode.main import QRCode
import mpv
import os

from .entry import Entry


class QRPosition(Enum):
    TOP_LEFT = 1
    TOP_RIGHT = 2
    BOTTOM_LEFT = 3
    BOTTOM_RIGHT = 4

    @staticmethod
    def from_string(value: str) -> "QRPosition":
        match value:
            case "top-left":
                return QRPosition.TOP_LEFT
            case "top-right":
                return QRPosition.TOP_RIGHT
            case "bottom-left":
                return QRPosition.BOTTOM_LEFT
            case "bottom-right":
                return QRPosition.BOTTOM_RIGHT
            case _:
                return QRPosition.BOTTOM_RIGHT


class Player:
    def __init__(
        self,
        config: dict[str, Any],
        quit_callback: Callable[[], None],
        queue: Optional[list[Entry]] = None,
    ) -> None:
        locale.setlocale(locale.LC_ALL, "C")
        qr_string = f"{config['server']}/{config['room']}"

        self.queue = queue if queue is not None else []
        self.base_dir = f"{os.path.dirname(__file__)}/static"
        if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
            self.base_dir = getattr(sys, "_MEIPASS")
        self.closing = False
        self.mpv: Optional[mpv.MPV] = None
        self.qr_overlay: Optional[mpv.ImageOverlay] = None
        self.qr_box_size = 1 if config["qr_box_size"] < 1 else config["qr_box_size"]
        self.qr_position = QRPosition.from_string(config["qr_position"])
        self.next_up_time = config.get("next_up_time", 20)
        self.update_qr(
            qr_string,
        )

        self.default_options = {
            "scale": "bilinear",
        }
        self.quit_callback = quit_callback
        self.callback_audio_load: Optional[str] = None

    def start(self) -> None:
        self.mpv = mpv.MPV(
            ytdl=True,
            input_default_bindings=True,
            input_vo_keyboard=True,
            osc=True,
            osd_border_style="background-box",
            osd_back_color="#E0008000",
            osd_color="#E0FFFFFF",
            osd_outline_color="#50000000",
            osd_shadow_offset=10,
            osd_align_x="center",
            osd_align_y="top",
        )
        self.next_up_overlay_id = self.mpv.allocate_overlay_id()
        self.next_up_y_pos = -120
        self.mpv.title = "Syng.Rocks! - Player"
        self.mpv.keep_open = "yes"
        self.mpv.play(
            f"{self.base_dir}/background.png",
        )
        self.mpv.observe_property("osd-width", self.osd_size_handler)
        self.mpv.observe_property("osd-height", self.osd_size_handler)
        self.mpv.observe_property("playtime-remaining", self.playtime_remaining_handler)
        self.mpv.register_event_callback(self.event_callback)

    def playtime_remaining_handler(self, attribute: str, value: float) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return
        hidden = value is None or value > self.next_up_time

        if len(self.queue) < 2:
            return
        if not hidden:
            if self.next_up_y_pos < 0:
                self.next_up_y_pos += 5
        else:
            self.next_up_y_pos = -120
        entry = self.queue[1]

        self.mpv.command(
            "osd_overlay",
            id=self.next_up_overlay_id,
            data=f"{{\\pos({1920 // 2},{self.next_up_y_pos})}}Next Up: {entry.artist} - {entry.title} ({entry.performer})",
            res_x=1920,
            res_y=1080,
            z=0,
            hidden=hidden,
            format="ass-events",
        )

    def event_callback(self, event: mpv.MpvEvent) -> None:
        e = event.as_dict()
        if e["event"] == b"shutdown":
            if not self.closing:
                self.closing = True
                self.quit_callback()
        elif e["event"] == b"file-loaded":
            if self.callback_audio_load is not None and self.mpv is not None:
                self.mpv.audio_add(self.callback_audio_load)
                self.callback_audio_load = None

    def update_qr(self, qr_string: str) -> None:
        qr = QRCode(box_size=self.qr_box_size, border=1)
        qr.add_data(qr_string)
        qr.make()
        self.qr = qr.make_image().convert("RGBA")

    def osd_size_handler(self, attribute: str, value: int) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return
        if self.qr_overlay:
            self.mpv.remove_overlay(self.qr_overlay.overlay_id)

        osd_width: int = cast(int, self.mpv.osd_width)
        osd_height: int = cast(int, self.mpv.osd_height)

        match self.qr_position:
            case QRPosition.BOTTOM_RIGHT:
                x_pos = osd_width - self.qr.width - 10
                y_pos = osd_height - self.qr.height - 10
            case QRPosition.BOTTOM_LEFT:
                x_pos = 10
                y_pos = osd_height - self.qr.height - 10
            case QRPosition.TOP_RIGHT:
                x_pos = osd_width - self.qr.width - 10
                y_pos = 10
            case QRPosition.TOP_LEFT:
                x_pos = 10
                y_pos = 10

        self.qr_overlay = self.mpv.create_image_overlay(self.qr, pos=(x_pos, y_pos))

    async def queue_next(self, entry: Entry) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return

        loop = asyncio.get_running_loop()

        frame = sys._getframe()
        stream_name = f"__python_mpv_play_generator_{hash(frame)}"

        @self.mpv.python_stream(stream_name)
        def preview() -> Iterable[bytes]:
            subtitle: str = f"""1
00:00:00,00 --> 00:05:00,00
{entry.artist} - {entry.title}
{entry.performer}"""
            yield subtitle.encode()
            preview.unregister()

        self.mpv.sub_pos = 50
        self.play_image(
            f"{self.base_dir}/background20perc.png", 3, sub_file=f"python://{stream_name}"
        )

        try:
            await loop.run_in_executor(None, self.mpv.wait_for_property, "eof-reached")
        except mpv.ShutdownError:
            self.quit_callback()

    def play_image(self, image: str, duration: int, sub_file: Optional[str] = None) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return

        for property, value in self.default_options.items():
            self.mpv[property] = value
        self.mpv.image_display_duration = duration
        self.mpv.keep_open = "yes"
        if sub_file:
            self.mpv.loadfile(image, sub_file=sub_file)
        else:
            self.mpv.loadfile(image)
        self.mpv.pause = False

    async def play(
        self,
        video: str,
        audio: Optional[str] = None,
        override_options: Optional[dict[str, str]] = None,
    ) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return

        if override_options is None:
            override_options = {}
        for property, value in self.default_options.items():
            self.mpv[property] = value

        for property, value in override_options.items():
            self.mpv[property] = value

        loop = asyncio.get_running_loop()
        self.mpv.pause = True
        if audio:
            self.callback_audio_load = audio
            self.mpv.loadfile(video)
        else:
            self.mpv.loadfile(video)
        self.mpv.pause = False
        try:
            await loop.run_in_executor(None, self.mpv.wait_for_property, "eof-reached")
            self.mpv.image_display_duration = 0
            self.mpv.play(f"{self.base_dir}/background.png")
        except mpv.ShutdownError:
            self.quit_callback()

    def skip_current(self) -> None:
        if self.mpv is None:
            print("MPV is not initialized", file=sys.stderr)
            return

        self.mpv.image_display_duration = 0
        self.mpv.play(
            f"{self.base_dir}/background.png",
        )



"""Module for an abstract filebased Source."""

import asyncio
import os
from typing import TYPE_CHECKING, Any, Optional

from syng.entry import Entry


try:
    from pymediainfo import MediaInfo

    PYMEDIAINFO_AVAILABLE = True
except ImportError:
    if TYPE_CHECKING:
        from pymediainfo import MediaInfo
    PYMEDIAINFO_AVAILABLE = False

from .source import Source
from ..config import ListStrOption, ConfigOption


class FileBasedSource(Source):
    """
    A abstract source for indexing and playing songs based on files.

    Config options are:
        -``extensions``, list of filename extensions
    """

    config_schema = Source.config_schema | {
        "extensions": ConfigOption(
            ListStrOption(),
            "List of filename extensions\n(mp3+cdg, mp4, ...)",
            ["mp3+cdg"],
        ),
    }

    def apply_config(self, config: dict[str, Any]) -> None:
        self.build_index = True
        self.extensions: list[str] = config["extensions"] if "extensions" in config else ["mp3+cdg"]
        self.extra_mpv_options = {"scale": "oversample"}

    def is_valid(self, entry: Entry) -> bool:
        return entry.ident in self._index and entry.source == self.source_name

    def has_correct_extension(self, path: Optional[str]) -> bool:
        """
        Check if a `path` has a correct extension.

        For A+B type extensions (like mp3+cdg) only the latter halve is checked

        :param path: The path to check.
        :type path: Optional[str]
        :return: True iff path has correct extension.
        :rtype: bool
        """
        return path is not None and os.path.splitext(path)[1][1:] in [
            ext.rsplit("+", maxsplit=1)[-1] for ext in self.extensions
        ]

    def get_video_audio_split(self, path: str) -> tuple[str, Optional[str]]:
        """
        Returns path for audio and video file, if filetype is marked as split.

        If the file is not marked as split, the second element of the tuple will be None.

        :params: path: The path to the file
        :type path: str
        :return: Tuple with path to video and audio file
        :rtype: tuple[str, Optional[str]]
        """
        extension_of_path = os.path.splitext(path)[1][1:]
        splitted_extensions = [ext.split("+") for ext in self.extensions if "+" in ext]
        splitted_extensions_dict = {video: audio for [audio, video] in splitted_extensions}

        if extension_of_path in splitted_extensions_dict:
            audio_path = (
                os.path.splitext(path)[0] + "." + splitted_extensions_dict[extension_of_path]
            )
            return (path, audio_path)
        return (path, None)

    async def get_duration(self, path: str) -> int:
        """
        Return the duration for the file.

        :param path: The path to the file
        :type path: str
        :return: The duration in seconds
        :rtype: int
        """
        if not PYMEDIAINFO_AVAILABLE:
            return 180

        def _get_duration(file: str) -> int:
            info: str | MediaInfo = MediaInfo.parse(file)
            if isinstance(info, str):
                return 180
            duration: int = info.audio_tracks[0].to_data()["duration"]
            return duration // 1000

        video_path, audio_path = self.get_video_audio_split(path)

        check_path = audio_path if audio_path is not None else video_path
        duration = await asyncio.to_thread(_get_duration, check_path)

        return duration



"""
Construct the YouTube source.

This source uses yt-dlp to search and download videos from YouTube.

Adds it to the ``available_sources`` with the name ``youtube``.
"""

from __future__ import annotations

import asyncio
import shlex
from functools import partial
from urllib.parse import urlencode
from typing import Any, Optional, Tuple

from yt_dlp import YoutubeDL
from yt_dlp.utils import DownloadError
from platformdirs import user_cache_dir


from ..entry import Entry
from ..result import Result
from .source import Source, available_sources
from ..config import (
    BoolOption,
    ChoiceOption,
    FolderOption,
    ListStrOption,
    ConfigOption,
    StrOption,
    IntOption,
)


class YouTube:
    """
    A minimal compatibility layer for the YouTube object of pytube, implemented via yt-dlp
    """

    def __init__(self, url: Optional[str] = None, info: Optional[dict[str, Any]] = None):
        """
        Construct a YouTube object from a url.

        If the url is already in the cache, the object is constructed from the
        cache. Otherwise yt-dlp is used to extract the information.

        :param url: The url of the video.
        :type url: Optional[str]
        """
        self._title: Optional[str]
        self._author: Optional[str]

        if url is not None:
            try:
                if info is not None:
                    self._infos = info
                else:
                    self._infos = YoutubeDL({"quiet": True}).extract_info(url, download=False)
            except DownloadError:
                self.length = 300
                self._title = None
                self._author = None
                self.watch_url = url
                return
            if self._infos is None:
                raise RuntimeError(f'Extraction not possible for "{url}"')
            self.length = int(self._infos["duration"])
            self._title = self._infos["title"]
            self._author = self._infos["channel"]
            self.watch_url = url
        else:
            self.length = 0
            self._title = ""
            self.channel = ""
            self._author = ""
            self.watch_url = ""

    @property
    def title(self) -> str:
        """
        The title of the video.

        :return: The title of the video.
        :rtype: str
        """
        if self._title is None:
            return ""
        return self._title

    @property
    def author(self) -> str:
        """
        The author of the video.

        :return: The author of the video.
        :rtype: str
        """
        if self._author is None:
            return ""
        return self._author

    @classmethod
    def from_result(cls, search_result: dict[str, Any]) -> YouTube:
        """
        Construct a YouTube object from yt-dlp search results.

        :param search_result: The search result from yt-dlp.
        :type search_result: dict[str, Any]
        """
        url = search_result["url"]
        return cls(url, info=search_result)


class Search:
    """
    A minimal compatibility layer for the Search object of pytube, implemented via yt-dlp
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, query: str, channel: Optional[str] = None):
        """
        Construct a Search object from a query and an optional channel.

        Uses yt-dlp to search for the query.

        If no channel is given, the search is done on the whole of YouTube.

        :param query: The query to search for.
        :type query: str
        :param channel: The channel to search in.
        :type channel: Optional[str]
        """
        sp = "EgIQAfABAQ=="  # This is a magic string, that tells youtube to search for videos
        if channel is None:
            query_url = (
                f"https://youtube.com/results?{urlencode({'search_query': query, 'sp': sp})}"
            )
        else:
            if channel[0] == "/":
                channel = channel[1:]
            query_url = (
                f"https://www.youtube.com/{channel}/search?{urlencode({'query': query, 'sp':sp})}"
            )

        results = YoutubeDL(
            {
                "extract_flat": True,
                "quiet": True,
                "playlist_items": ",".join(map(str, range(1, 51))),
            }
        ).extract_info(
            query_url,
            download=False,
        )
        self.results = []
        if results is not None:
            filtered_entries = filter(lambda entry: "short" not in entry["url"], results["entries"])

            for r in filtered_entries:
                try:
                    self.results.append(YouTube.from_result(r))
                except KeyError:
                    pass


class YoutubeSource(Source):
    """A source for playing karaoke files from YouTube.

    Config options are:
        - ``channels``: A list of all channel this source should search in.
          Examples are ``/c/CCKaraoke`` or
          ``/channel/UCwTRjvjVge51X-ILJ4i22ew``
        - ``tmp_dir``: The folder, where temporary files are stored. Default
          is ``${XDG_CACHE_DIR}/syng``.
        - ``max_res``: The highest video resolution, that should be
          downloaded/streamed. Default is 720.
        - ``start_streaming``: If set to ``True``, the client starts streaming
          the video, if buffering was not completed. Needs ``youtube-dl`` or
          ``yt-dlp``. Default is False.
        - ``search_suffix``: A string that is appended to the search query.
          Default is "karaoke".
        - ``max_duration``: The maximum duration of a video in seconds. A value of 0 disables this. Default is 1800.
    """

    source_name = "youtube"
    config_schema = Source.config_schema | {
        "enabled": ConfigOption(BoolOption(), "Enable this source", True),
        "channels": ConfigOption(
            ListStrOption(), "A list channels\nto search in", [], send_to_server=True
        ),
        "tmp_dir": ConfigOption(
            FolderOption(), "Folder for\ntemporary download", user_cache_dir("syng")
        ),
        "max_res": ConfigOption(
            ChoiceOption(["144", "240", "360", "480", "720", "1080", "2160"]),
            "Maximum resolution\nto download",
            "720",
        ),
        "start_streaming": ConfigOption(
            BoolOption(),
            "Start streaming if\ndownload is not complete",
            False,
        ),
        "search_suffix": ConfigOption(
            StrOption(),
            "A string that is appended\nto each search query",
            "karaoke",
            send_to_server=True,
        ),
        "max_duration": ConfigOption(
            IntOption(),
            "The maximum duration\nof a video in seconds\nA value of 0 disables this",
            1800,
            send_to_server=True,
        ),
    }

    def apply_config(self, config: dict[str, Any]) -> None:
        self.channels: list[str] = config["channels"] if "channels" in config else []
        self.tmp_dir: str = config["tmp_dir"] if "tmp_dir" in config else "/tmp/syng"
        try:
            self.max_res: int = int(config["max_res"])
        except (ValueError, KeyError):
            self.max_res = 720
        self.start_streaming: bool = (
            config["start_streaming"] if "start_streaming" in config else False
        )
        self.formatstring = (
            f"bestvideo[height<={self.max_res}]+" f"bestaudio/best[height<={self.max_res}]"
        )
        self.search_suffix = config.get("search_suffix", "karaoke")
        self.extra_mpv_options = {"ytdl-format": self.formatstring}
        self._yt_dlp = YoutubeDL(
            params={
                "paths": {"home": self.tmp_dir},
                "format": self.formatstring,
                "quiet": True,
            }
        )
        self.max_duration: int = config.get("max_duration", 1800)

    async def ensure_playable(self, entry: Entry) -> tuple[str, Optional[str]]:
        """
        Ensure that the entry is playable.

        If the entry is not yet downloaded, download it.
        If start_streaming is set, start streaming immediatly.

        :param entry: The entry to download.
        :type entry: Entry
        :rtype: None
        """

        if entry.incomplete_data:
            meta_info = await self.get_missing_metadata(entry)
            entry.update(**meta_info)

        if self.max_duration > 0 and entry.duration > self.max_duration:
            raise ValueError(f"Video {entry.ident} too long.")

        if self.start_streaming and not self.downloaded_files[entry.ident].complete:
            return (entry.ident, None)

        return await super().ensure_playable(entry)

    async def get_entry(
        self,
        performer: str,
        ident: str,
        /,
        artist: Optional[str] = None,
        title: Optional[str] = None,
    ) -> Optional[Entry]:
        """
        Create an :py:class:`syng.entry.Entry` for the identifier.

        The identifier should be a youtube url. An entry is created with
        all available metadata for the video.

        :param performer: The person singing.
        :type performer: str
        :param ident: A url to a YouTube video.
        :type ident: str
        :return: An entry with the data.
        :rtype: Optional[Entry]
        """

        return Entry(
            ident=ident,
            source="youtube",
            duration=180,
            album="YouTube",
            title=title,
            artist=artist,
            performer=performer,
            incomplete_data=True,
        )

    async def search(self, query: str) -> list[Result]:
        """
        Search YouTube and the configured channels for the query.

        The first results are the results of the configured channels. The next
        results are the results from youtube as a whole, a configurable suffix
        is appended to the search query (default is "karaoke").

        All results are sorted by how good they match to the search query,
        respecting their original source (channel or YouTube as a whole).

        All searching is done concurrently.

        :param query: The query to search for
        :type query: str
        :return: A list of Results.
        :rtype: list[Result]
        """

        def _contains_index(query: str, result: YouTube) -> float:
            """
            Calculate a score for the result.

            The score is the ratio of how many words of the query are in the
            title and author of the result.

            :param query: The query to search for.
            :type query: str
            :param result: The result to score.
            :type result: YouTube
            """
            compare_string: str = result.title.lower() + " " + result.author.lower()
            hits: int = 0
            queries: list[str] = shlex.split(query.lower())
            for word in queries:
                if word in compare_string:
                    hits += 1

            return 1 - (hits / len(queries))

        results: list[YouTube] = []
        results_lists: list[list[YouTube]] = await asyncio.gather(
            *[asyncio.to_thread(self._channel_search, query, channel) for channel in self.channels],
            asyncio.to_thread(self._yt_search, query),
        )
        results = [search_result for yt_result in results_lists for search_result in yt_result]

        results.sort(key=partial(_contains_index, query))

        return [
            Result(
                ident=result.watch_url,
                source="youtube",
                title=result.title,
                artist=result.author,
                album="YouTube",
                duration=str(result.length),
            )
            for result in results
            if self.max_duration == 0 or result.length <= self.max_duration
        ]

    def is_valid(self, entry: Entry) -> bool:
        """
        Check if the entry is valid.

        An entry is valid, if the video is not too long.

        :param entry: The entry to check.
        :type entry: Entry
        :return: True if the entry is valid, False otherwise.
        :rtype: bool
        """
        return self.max_duration == 0 or entry.duration <= self.max_duration

    def _yt_search(self, query: str) -> list[YouTube]:
        """Search youtube as a whole.

        Adds a configurable suffix to the query. Default is "karaoke".
        """
        suffix = f" {self.search_suffix}" if self.search_suffix else ""
        return Search(f"{query}{suffix}").results

    def _channel_search(self, query: str, channel: str) -> list[YouTube]:
        """
        Search a channel for a query.

        A lot of black Magic happens here.
        """
        return Search(f"{query} karaoke", channel).results

    async def get_missing_metadata(self, entry: Entry) -> dict[str, Any]:
        """
        Video metadata should be read on the client to avoid banning
        the server.
        """
        if entry.incomplete_data or None in (entry.artist, entry.title):
            youtube_video: YouTube = await asyncio.to_thread(YouTube, entry.ident)
            return {
                "duration": youtube_video.length,
                "artist": youtube_video.author,
                "title": youtube_video.title,
            }
        return {}

    async def do_buffer(self, entry: Entry, pos: int) -> Tuple[str, Optional[str]]:
        """
        Download the video.

        Downloads the highest quality stream respecting the ``max_res``.
        For higher resolution videos (1080p and above).

        Yt-dlp automatically merges the audio and video, so only the video
        location exists, the return value for the audio part will always be
        ``None``.

        If pos is 0 and start_streaming is set, no buffering is done, instead the
        youtube url is returned.

        :param entry: The entry to download.
        :type entry: Entry
        :param pos: The position in the video to start buffering.
        :type pos: int
        :return: The location of the video file and ``None``.
        :rtype: Tuple[str, Optional[str]]
        """

        if self.max_duration > 0 and entry.duration > self.max_duration:
            raise ValueError(
                f"Video {entry.ident} too long: {entry.duration} > {self.max_duration}"
            )

        if pos == 0 and self.start_streaming:
            return entry.ident, None

        info: Any = await asyncio.to_thread(self._yt_dlp.extract_info, entry.ident)
        combined_path = info["requested_downloads"][0]["filepath"]
        return combined_path, None


available_sources["youtube"] = YoutubeSource



import asyncio
from io import BytesIO
import sys
import logging
from logging.handlers import QueueListener
from logging.handlers import QueueHandler
import packaging.version

import aiohttp
from queue import Queue
from collections.abc import Callable
from datetime import datetime
import os
from functools import partial
import random
from typing import TYPE_CHECKING, Any, Optional
import secrets
import string
import signal


try:
    if not TYPE_CHECKING:
        from ctypes import windll

        appid = "rocks.syng.Syng.2.1.0"
        windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)
except ImportError:
    pass

os.environ["QT_API"] = "pyqt6"
from qasync import QEventLoop, QApplication
from PyQt6.QtCore import (
    QAbstractListModel,
    QModelIndex,
    QObject,
    QTimer,
    Qt,
    pyqtSignal,
    pyqtSlot,
)
from PyQt6.QtGui import QCloseEvent, QIcon, QPixmap
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDateTimeEdit,
    QFileDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QListView,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QTabBar,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from yaml import dump, load, Loader, Dumper
from qrcode.main import QRCode
import platformdirs

from . import __version__, resources  # noqa
from .client import Client, default_config
from .log import logger
from .entry import Entry

from .sources import available_sources
from .config import (
    BoolOption,
    ChoiceOption,
    FileOption,
    FolderOption,
    IntOption,
    ListStrOption,
    PasswordOption,
    StrOption,
)


class QueueModel(QAbstractListModel):
    def __init__(self, queue: list[Entry]) -> None:
        super().__init__()
        self.queue = queue

    def update(self, queue: list[Entry]) -> None:
        self.queue = queue
        self.dataChanged.emit(self.index(0, 0), self.index(self.rowCount() - 1, 0))

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            entry = self.queue[index.row()]
            return f"{entry.title} - {entry.artist} [{entry.album}]\n{entry.performer}"

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.queue)


class QueueView(QListView):
    pass


class OptionFrame(QWidget):
    def add_bool_option(self, name: str, description: str, value: bool = False) -> None:
        label = QLabel(description, self)

        self.bool_options[name] = QCheckBox(self)
        self.bool_options[name].setChecked(value)
        self.form_layout.addRow(label, self.bool_options[name])

    def add_string_option(
        self,
        name: str,
        description: str,
        value: Optional[str] = "",
        callback: Optional[Callable[..., None]] = None,
        is_password: bool = False,
    ) -> None:
        if value is None:
            value = ""

        label = QLabel(description, self)

        self.string_options[name] = QLineEdit(self)
        if is_password:
            self.string_options[name].setEchoMode(QLineEdit.EchoMode.Password)
            action = self.string_options[name].addAction(
                QIcon(":/icons/eye_strike.svg"),
                QLineEdit.ActionPosition.TrailingPosition,
            )

            if action is not None:

                def toggle_visibility() -> None:
                    self.string_options[name].setEchoMode(
                        QLineEdit.EchoMode.Normal
                        if self.string_options[name].echoMode() == QLineEdit.EchoMode.Password
                        else QLineEdit.EchoMode.Password
                    )
                    if self.string_options[name].echoMode() == QLineEdit.EchoMode.Password:
                        action.setIcon(QIcon(":/icons/eye_strike.svg"))
                    else:
                        action.setIcon(QIcon(":/icons/eye_clear.svg"))

                action.triggered.connect(toggle_visibility)

        self.string_options[name].insert(value)
        self.form_layout.addRow(label, self.string_options[name])
        self.rows[name] = (label, self.string_options[name])
        if callback is not None:
            self.string_options[name].textChanged.connect(callback)

    def path_setter(self, line: QLineEdit, name: Optional[str]) -> None:
        if name:
            line.setText(name)

    def add_file_option(
        self,
        name: str,
        description: str,
        value: Optional[str] = "",
        callback: Optional[Callable[..., None]] = None,
    ) -> None:
        if value is None:
            value = ""

        label = QLabel(description, self)
        file_layout = QHBoxLayout()
        file_name_widget = QLineEdit(value, self)
        file_button = QPushButton(QIcon.fromTheme("document-open"), "", self)

        file_button.clicked.connect(
            lambda: self.path_setter(
                file_name_widget,
                QFileDialog.getOpenFileName(
                    self, "Select File", directory=os.path.dirname(file_name_widget.text())
                )[0],
            )
        )

        if callback is not None:
            file_name_widget.textChanged.connect(callback)

        file_layout.addWidget(file_name_widget)
        file_layout.addWidget(file_button)

        self.string_options[name] = file_name_widget
        self.rows[name] = (label, file_name_widget)
        self.form_layout.addRow(label, file_layout)

    def add_folder_option(
        self,
        name: str,
        description: str,
        value: Optional[str] = "",
        callback: Optional[Callable[..., None]] = None,
    ) -> None:
        if value is None:
            value = ""

        label = QLabel(description, self)
        folder_layout = QHBoxLayout()
        folder_name_widget = QLineEdit(value, self)
        folder_button = QPushButton(QIcon.fromTheme("folder-open"), "", self)

        folder_button.clicked.connect(
            lambda: self.path_setter(
                folder_name_widget,
                QFileDialog.getExistingDirectory(
                    self, "Select Folder", directory=folder_name_widget.text()
                ),
            )
        )

        if callback is not None:
            folder_name_widget.textChanged.connect(callback)

        folder_layout.addWidget(folder_name_widget)
        folder_layout.addWidget(folder_button)

        self.string_options[name] = folder_name_widget
        self.rows[name] = (label, folder_name_widget)
        self.form_layout.addRow(label, folder_layout)

    def add_int_option(
        self,
        name: str,
        description: str,
        value: Optional[int] = 0,
        callback: Optional[Callable[..., None]] = None,
    ) -> None:
        if value is None:
            value = 0

        label = QLabel(description, self)

        self.int_options[name] = QSpinBox(self)
        self.int_options[name].setMaximum(9999)
        self.int_options[name].setValue(value)
        self.form_layout.addRow(label, self.int_options[name])
        self.rows[name] = (label, self.int_options[name])
        if callback is not None:
            self.int_options[name].textChanged.connect(callback)

    def del_list_element(
        self,
        name: str,
        element: QLineEdit,
        line: QWidget,
        layout: QVBoxLayout,
    ) -> None:
        self.list_options[name].remove(element)

        layout.removeWidget(line)
        line.deleteLater()

    def add_list_element(
        self,
        name: str,
        layout: QVBoxLayout,
        init: str,
        callback: Optional[Callable[..., None]],
    ) -> None:
        input_and_minus = QWidget()
        input_and_minus_layout = QHBoxLayout(input_and_minus)
        input_and_minus.setLayout(input_and_minus_layout)

        input_and_minus_layout.setContentsMargins(0, 0, 0, 0)

        input_field = QLineEdit(input_and_minus)
        input_field.setText(init)
        input_field.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        input_and_minus_layout.addWidget(input_field)
        if callback is not None:
            input_field.textChanged.connect(callback)

        minus_button = QPushButton(QIcon.fromTheme("list-remove"), "", input_and_minus)
        minus_button.clicked.connect(
            partial(self.del_list_element, name, input_field, input_and_minus, layout)
        )
        minus_button.setFixedWidth(40)
        minus_button.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        input_and_minus_layout.addWidget(minus_button)

        layout.insertWidget(layout.count() - 1, input_and_minus)

        self.list_options[name].append(input_field)

    def add_list_option(
        self,
        name: str,
        description: str,
        value: list[str],
        callback: Optional[Callable[..., None]] = None,
    ) -> None:
        label = QLabel(description, self)

        container_layout = QVBoxLayout()

        self.form_layout.addRow(label, container_layout)
        self.rows[name] = (label, container_layout)

        self.list_options[name] = []
        for v in value:
            self.add_list_element(name, container_layout, v, callback)
        plus_button = QPushButton(QIcon.fromTheme("list-add"), "", self)
        plus_button.clicked.connect(
            partial(self.add_list_element, name, container_layout, "", callback)
        )
        plus_button.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        container_layout.addWidget(plus_button)

    def add_choose_option(
        self, name: str, description: str, values: list[str], value: str = ""
    ) -> None:
        label = QLabel(description, self)

        self.choose_options[name] = QComboBox(self)
        self.choose_options[name].addItems(values)
        self.choose_options[name].setCurrentText(str(value))
        self.form_layout.addRow(label, self.choose_options[name])
        self.rows[name] = (label, self.choose_options[name])

    def add_date_time_option(self, name: str, description: str, value: str) -> None:
        label = QLabel(description, self)
        date_time_layout = QHBoxLayout()
        date_time_widget = QDateTimeEdit(self)
        date_time_enabled = QCheckBox("Enabled", self)
        date_time_enabled.stateChanged.connect(
            lambda: date_time_widget.setEnabled(date_time_enabled.isChecked())
        )

        self.date_time_options[name] = (date_time_widget, date_time_enabled)
        date_time_widget.setCalendarPopup(True)
        try:
            date_time_widget.setDateTime(datetime.fromisoformat(value))
            date_time_enabled.setChecked(True)
        except (TypeError, ValueError):
            date_time_widget.setDateTime(datetime.now())
            date_time_widget.setEnabled(False)
            date_time_enabled.setChecked(False)

        date_time_layout.addWidget(date_time_widget)
        date_time_layout.addWidget(date_time_enabled)

        self.form_layout.addRow(label, date_time_layout)
        self.rows[name] = (label, date_time_layout)

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.form_layout = QFormLayout(self)
        self.setLayout(self.form_layout)
        self.string_options: dict[str, QLineEdit] = {}
        self.int_options: dict[str, QSpinBox] = {}
        self.choose_options: dict[str, QComboBox] = {}
        self.bool_options: dict[str, QCheckBox] = {}
        self.list_options: dict[str, list[QLineEdit]] = {}
        self.date_time_options: dict[str, tuple[QDateTimeEdit, QCheckBox]] = {}
        self.rows: dict[str, tuple[QLabel, QWidget | QLayout]] = {}

    @property
    def option_names(self) -> set[str]:
        return set(
            self.string_options.keys()
            | self.int_options.keys()
            | self.choose_options.keys()
            | self.bool_options.keys()
            | self.list_options.keys()
            | self.date_time_options.keys()
        )

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = {}
        for name, textbox in self.string_options.items():
            config[name] = textbox.text().strip()

        for name, spinner in self.int_options.items():
            config[name] = spinner.value()

        for name, optionmenu in self.choose_options.items():
            config[name] = optionmenu.currentText().strip()

        for name, checkbox in self.bool_options.items():
            config[name] = checkbox.isChecked()

        for name, textboxes in self.list_options.items():
            config[name] = []
            for textbox in textboxes:
                config[name].append(textbox.text().strip())

        for name, (picker, checkbox) in self.date_time_options.items():
            if not checkbox.isChecked():
                config[name] = None
                continue
            try:
                config[name] = picker.dateTime().toPyDateTime().isoformat()
            except ValueError:
                config[name] = None

        return config

    def load_config(self, config: dict[str, Any]) -> None:
        for name, textbox in self.string_options.items():
            textbox.setText(config[name])

        for name, spinner in self.int_options.items():
            try:
                spinner.setValue(config[name])
            except ValueError:
                spinner.setValue(0)

        for name, optionmenu in self.choose_options.items():
            optionmenu.setCurrentText(str(config[name]))

        for name, checkbox in self.bool_options.items():
            checkbox.setChecked(config[name])

        for name, textboxes in self.list_options.items():
            for i, textbox in enumerate(textboxes):
                textbox.setText(config[name][i])

        for name, (picker, checkbox) in self.date_time_options.items():
            if config[name] is not None:
                picker.setDateTime(datetime.fromisoformat(config[name]))
                checkbox.setChecked(True)
            else:
                picker.setDateTime(datetime.now())
                picker.setEnabled(False)
                checkbox.setChecked(False)


class SourceTab(OptionFrame):
    def __init__(self, parent: QWidget, source_name: str, config: dict[str, Any]) -> None:
        super().__init__(parent)
        source = available_sources[source_name]
        self.vars: dict[str, str | bool | list[str]] = {}
        for name, option in source.config_schema.items():
            value = config[name] if name in config else option.default
            match option.type:
                case BoolOption():
                    self.add_bool_option(name, option.description, value=value)
                case ListStrOption():
                    self.add_list_option(name, option.description, value=value)
                case StrOption():
                    self.add_string_option(name, option.description, value=value)
                case IntOption():
                    self.add_int_option(name, option.description, value=value)
                case PasswordOption():
                    self.add_string_option(name, option.description, value=value, is_password=True)
                case FolderOption():
                    self.add_folder_option(name, option.description, value=value)
                case FileOption():
                    self.add_file_option(name, option.description, value=value)
                case ChoiceOption(choices):
                    self.add_choose_option(name, option.description, choices, value)


class GeneralConfig(OptionFrame):
    def __init__(
        self,
        parent: QWidget,
        config: dict[str, Any],
        callback: Callable[..., None],
    ) -> None:
        super().__init__(parent)

        self.add_string_option("server", "Server", config["server"], callback)
        self.add_string_option("room", "Room", config["room"], callback)
        self.add_string_option("secret", "Admin Password", config["secret"], is_password=True)
        self.add_choose_option(
            "waiting_room_policy",
            "Waiting room policy",
            ["forced", "optional", "none"],
            str(config["waiting_room_policy"]).lower(),
        )
        self.add_date_time_option("last_song", "Last song ends at", config["last_song"])
        self.add_int_option(
            "preview_duration", "Preview duration in seconds", int(config["preview_duration"])
        )
        self.add_int_option(
            "next_up_time",
            "Time remaining before Next Up Box is shown",
            int(config["next_up_time"]),
        )
        self.add_string_option(
            "key", "Key for server (if necessary)", config["key"], is_password=True
        )
        self.add_int_option(
            "buffer_in_advance",
            "Buffer the next songs in advance",
            int(config["buffer_in_advance"]),
        )
        self.add_int_option("qr_box_size", "QR Code Box Size", int(config["qr_box_size"]))
        self.add_choose_option(
            "qr_position",
            "QR Code Position",
            ["top-left", "top-right", "bottom-left", "bottom-right"],
            config["qr_position"],
        )
        self.add_choose_option(
            "log_level",
            "Log Level",
            ["debug", "info", "warning", "error", "critical"],
            config["log_level"],
        )

        self.simple_options = ["server", "room", "secret"]

        if not config["show_advanced"]:
            for option in self.option_names.difference(self.simple_options):
                self.rows[option][0].setVisible(False)
                widget_or_layout = self.rows[option][1]
                if isinstance(widget_or_layout, QWidget):
                    widget_or_layout.setVisible(False)
                else:
                    for i in range(widget_or_layout.count()):
                        item = widget_or_layout.itemAt(i)
                        widget = item.widget() if item else None
                        if widget:
                            widget.setVisible(False)

    def get_config(self) -> dict[str, Any]:
        config = super().get_config()
        return config


class SyngGui(QMainWindow):
    def closeEvent(self, a0: Optional[QCloseEvent]) -> None:
        if self.client is not None:
            if self.client.player is not None and self.client.player.mpv is not None:
                self.client.player.mpv.terminate()
            self.client.quit_callback()

        self.log_label_handler.cleanup()

        self.destroy()
        sys.exit(0)

    def add_buttons(self, show_advanced: bool) -> None:
        self.buttons_layout = QHBoxLayout()
        self.central_layout.addLayout(self.buttons_layout)

        self.resetbutton = QPushButton("Set Config to Default")
        self.exportbutton = QPushButton("Export Config")
        self.importbutton = QPushButton("Import Config")
        self.buttons_layout.addWidget(self.resetbutton)
        self.buttons_layout.addWidget(self.exportbutton)
        self.buttons_layout.addWidget(self.importbutton)
        self.resetbutton.clicked.connect(self.clear_config)
        self.exportbutton.clicked.connect(self.export_config)
        self.importbutton.clicked.connect(self.import_config)
        if not show_advanced:
            self.resetbutton.hide()
            self.exportbutton.hide()
            self.importbutton.hide()

        self.show_advanced_toggle = QCheckBox("Show Advanced Options")
        self.show_advanced_toggle.setChecked(show_advanced)
        self.show_advanced_toggle.stateChanged.connect(self.toggle_advanced)
        self.buttons_layout.addWidget(self.show_advanced_toggle)

        spacer_item = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.buttons_layout.addItem(spacer_item)

        if os.getenv("SYNG_DEBUG", "0") == "1":
            self.print_background_tasks_button = QPushButton("Print Background Tasks")
            self.print_background_tasks_button.clicked.connect(
                lambda: print(asyncio.all_tasks(self.loop))
            )
            self.buttons_layout.addWidget(self.print_background_tasks_button)

        self.startbutton = QPushButton("Connect")

        self.startbutton.clicked.connect(self.start_syng_client)
        self.buttons_layout.addWidget(self.startbutton)

    def export_queue(self) -> None:
        if self.client is not None:
            filename = QFileDialog.getSaveFileName(self, "Export Queue", "", "JSON Files (*.json)")[
                0
            ]
            if filename:
                self.client.export_queue(filename)
        else:
            QMessageBox.warning(
                self,
                "No Client Running",
                "You need to start the client before you can export the queue.",
            )

    def import_queue(self) -> None:
        if self.client is not None:
            filename = QFileDialog.getOpenFileName(self, "Import Queue", "", "JSON Files (*.json)")[
                0
            ]
            if filename:
                asyncio.create_task(self.client.import_queue(filename))
        else:
            QMessageBox.warning(
                self,
                "No Client Running",
                "You need to start the client before you can import a queue.",
            )

    def clear_cache(self) -> None:
        """
        Clear the cache directory of the client.
        """

        cache_dir = platformdirs.user_cache_dir("syng")
        if os.path.exists(cache_dir):
            answer = QMessageBox.question(
                self,
                "Clear Cache",
                f"Are you sure you want to clear the cache directory at {cache_dir}?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if answer == QMessageBox.StandardButton.Yes:
                for root, dirs, files in os.walk(cache_dir, topdown=False):
                    for name in files:
                        os.remove(os.path.join(root, name))
                    for name in dirs:
                        os.rmdir(os.path.join(root, name))
                QMessageBox.information(self, "Cache Cleared", "The cache has been cleared.")

    def remove_room(self) -> None:
        if self.client is not None:
            answer = QMessageBox.question(
                self,
                "Remove Room",
                "Are you sure you want to remove the room on the server? This will disconnect all clients and clear the queue.",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if answer == QMessageBox.StandardButton.Yes:
                asyncio.create_task(self.client.remove_room())
        else:
            QMessageBox.warning(
                self,
                "No Client Running",
                "You need to start the client before you can remove a room.",
            )

    def toggle_advanced(self, state: bool) -> None:
        self.resetbutton.setVisible(state)
        self.exportbutton.setVisible(state)
        self.importbutton.setVisible(state)

        for option in self.general_config.option_names.difference(
            self.general_config.simple_options
        ):
            self.general_config.rows[option][0].setVisible(state)
            widget_or_layout = self.general_config.rows[option][1]
            if isinstance(widget_or_layout, QWidget):
                widget_or_layout.setVisible(state)
            else:
                for i in range(widget_or_layout.count()):
                    item = widget_or_layout.itemAt(i)
                    widget = item.widget() if item else None
                    if widget:
                        widget.setVisible(state)

        tabbar: Optional[QTabBar] = self.tabview.tabBar()
        if not state:
            if tabbar is not None:
                tabbar.hide()
            self.tabview.setCurrentIndex(0)
            self.general_config.form_layout.addRow(self.qr_widget)
        else:
            if tabbar is not None:
                tabbar.show()
            self.frm.addWidget(self.qr_widget)

    def init_frame(self) -> None:
        self.frm = QHBoxLayout()
        self.central_layout.addLayout(self.frm)

    def init_tabs(self, show_advanced: bool) -> None:
        self.tabview: QTabWidget = QTabWidget(parent=self.central_widget)
        self.tabview.setAcceptDrops(False)
        self.tabview.setTabPosition(QTabWidget.TabPosition.West)
        self.tabview.setTabShape(QTabWidget.TabShape.Rounded)
        self.tabview.setDocumentMode(False)
        self.tabview.setTabsClosable(False)
        self.tabview.setObjectName("tabWidget")

        self.tabview.setTabText(0, "General")
        for i, source in enumerate(available_sources):
            self.tabview.setTabText(i + 1, source)

        if not show_advanced:
            tabbar = self.tabview.tabBar()
            if tabbar is not None:
                tabbar.hide()

        self.frm.addWidget(self.tabview)

    def add_qr(self, show_advanced: bool) -> None:
        self.qr_widget: QWidget = QWidget(parent=self.central_widget)
        self.qr_layout = QVBoxLayout(self.qr_widget)
        self.qr_widget.setLayout(self.qr_layout)

        self.qr_label = QLabel(self.qr_widget)
        self.linklabel = QLabel(self.qr_widget)

        self.qr_layout.addWidget(self.qr_label)
        self.qr_layout.addWidget(self.linklabel)
        self.qr_layout.setAlignment(self.linklabel, Qt.AlignmentFlag.AlignCenter)
        self.qr_layout.setAlignment(self.qr_label, Qt.AlignmentFlag.AlignCenter)

        self.linklabel.setOpenExternalLinks(True)
        if not show_advanced:
            self.general_config.form_layout.addRow(self.qr_widget)
        else:
            self.frm.addWidget(self.qr_widget)

    def add_general_config(self, config: dict[str, Any]) -> None:
        self.general_config = GeneralConfig(self, config, self.update_qr)
        self.tabview.addTab(self.general_config, "General")

    def add_source_config(self, source_name: str, source_config: dict[str, Any]) -> None:
        self.tabs[source_name] = SourceTab(self, source_name, source_config)
        self.tabview.addTab(self.tabs[source_name], source_name)

    def add_log_tab(self) -> None:
        self.log_tab = QWidget(parent=self.central_widget)
        self.log_layout = QVBoxLayout(self.log_tab)
        self.log_tab.setLayout(self.log_layout)

        self.log_text = QTextEdit(self.log_tab)
        self.log_text.setReadOnly(True)
        self.log_layout.addWidget(self.log_text)

        self.tabview.addTab(self.log_tab, "Logs")

    def add_queue_tab(self) -> None:
        self.queue_tab = QWidget(parent=self.central_widget)
        self.queue_layout = QVBoxLayout(self.queue_tab)
        self.queue_tab.setLayout(self.queue_layout)

        self.queue_list_view: QueueView = QueueView(self.queue_tab)
        self.queue_layout.addWidget(self.queue_list_view)

        self.tabview.addTab(self.queue_tab, "Queue")

    def add_admin_tab(self) -> None:
        self.admin_tab = QWidget(parent=self.central_widget)
        self.admin_layout = QVBoxLayout(self.admin_tab)
        self.admin_layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)
        self.admin_tab.setLayout(self.admin_layout)

        self.remove_room_button = QPushButton("Remove Room", self.admin_tab)
        self.remove_room_button.clicked.connect(self.remove_room)
        self.admin_layout.addWidget(self.remove_room_button)
        self.remove_room_button.setDisabled(True)

        self.export_queue_button = QPushButton("Export Queue", self.admin_tab)
        self.export_queue_button.clicked.connect(self.export_queue)
        self.admin_layout.addWidget(self.export_queue_button)
        self.export_queue_button.setDisabled(True)

        self.import_queue_button = QPushButton("Import Queue", self.admin_tab)
        self.import_queue_button.clicked.connect(self.import_queue)
        self.admin_layout.addWidget(self.import_queue_button)
        self.import_queue_button.setDisabled(True)

        self.update_config_button = QPushButton("Update Config")
        self.update_config_button.clicked.connect(self.update_config)
        self.admin_layout.addWidget(self.update_config_button)
        self.update_config_button.setDisabled(True)

        self.clear_cache_button = QPushButton("Clear Cache", self.admin_tab)
        self.clear_cache_button.clicked.connect(self.clear_cache)
        self.admin_layout.addWidget(self.clear_cache_button)

        self.version_label = QLabel(
            "",
            self.admin_tab,
        )
        self.admin_layout.addWidget(self.version_label)

        self.tabview.addTab(self.admin_tab, "Admin")

    def update_version_label(
        self,
        running_version: Optional[packaging.version.Version],
        current_version: Optional[packaging.version.Version],
    ) -> None:
        label_string = f"<i>Running version: {running_version}</i><br />Current version on pypi: {current_version}"
        if current_version is not None and running_version is not None:
            if current_version > running_version:
                label_string += '<br /><span style="color:red;">A new version is available! Please update syng.</span>'
            else:
                label_string += '<br /><span style="color:green;">You are running the latest version.</span><br />Visit <a href="https://site.syng.rocks/">syng.rocks</a> for more information.'
        self.version_label.setText(label_string)

    async def get_pypi_version(self) -> None:
        print("search")
        async with aiohttp.ClientSession() as session:
            async with session.get("https://pypi.org/pypi/syng/json") as resp:
                if resp.status == 200:
                    data = await resp.json()
                    versions = filter(
                        lambda v: not v.is_prerelease,
                        map(packaging.version.parse, data["releases"].keys()),
                    )
                    self.update_version_label(
                        running_version=packaging.version.parse(__version__),
                        current_version=max(versions),
                    )
        return None

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("Syng.Rocks!")

        if os.name != "nt":
            self.setWindowIcon(QIcon(":/icons/syng.ico"))

        self.loop = asyncio.get_event_loop()

        self.pypi_version: Optional[str] = None
        asyncio.run_coroutine_threadsafe(self.get_pypi_version(), self.loop)

        self.client: Optional[Client] = None
        self.syng_client_logging_listener: Optional[QueueListener] = None

        self.configfile = os.path.join(platformdirs.user_config_dir("syng"), "config.yaml")

        self.central_widget = QWidget(parent=self)
        self.central_layout = QVBoxLayout(self.central_widget)

        config = self.load_config(self.configfile)

        self.init_frame()
        self.init_tabs(config["config"]["show_advanced"])
        self.add_buttons(config["config"]["show_advanced"])
        self.add_general_config(config["config"])
        self.add_qr(config["config"]["show_advanced"])
        self.tabs: dict[str, SourceTab] = {}

        for source_name in available_sources:
            self.add_source_config(source_name, config["sources"][source_name])

        # self.add_queue_tab()
        self.add_admin_tab()
        self.add_log_tab()

        self.update_qr()

        self.logqueue: Queue[logging.LogRecord] = Queue()
        logger.addHandler(QueueHandler(self.logqueue))
        self.log_label_handler = LoggingLabelHandler(self)
        self.log_label_handler.log_signal_emiter.log_signal.connect(self.print_log)

        self.syng_client_logging_listener = QueueListener(self.logqueue, self.log_label_handler)
        self.syng_client_logging_listener.start()

        self.setCentralWidget(self.central_widget)

        self.timer = QTimer()
        self.timer.timeout.connect(self.check_if_client_is_running)

    def complete_config(self, config: dict[str, Any]) -> dict[str, Any]:
        output: dict[str, dict[str, Any]] = {"sources": {}, "config": default_config()}

        try:
            output["config"] |= config["config"]
        except (KeyError, TypeError):
            print("Could not load config")

        if not output["config"]["secret"]:
            output["config"]["secret"] = "".join(
                secrets.choice(string.ascii_letters + string.digits) for _ in range(8)
            )

        if output["config"]["room"] == "":
            output["config"]["room"] = "".join(
                [random.choice(string.ascii_letters) for _ in range(6)]
            ).upper()

        for source_name, source in available_sources.items():
            source_config = {}
            for name, option in source.config_schema.items():
                source_config[name] = option.default

            output["sources"][source_name] = source_config

            try:
                output["sources"][source_name] |= config["sources"][source_name]
            except (KeyError, TypeError):
                pass

        return output

    def clear_config(self) -> None:
        answer = QMessageBox.question(
            self,
            "Set to Config to Default",
            "Are you sure you want to clear the config?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if answer == QMessageBox.StandardButton.Yes:
            self.update_config(self.complete_config({"config": {}, "sources": {}}))

    def load_config(self, filename: str) -> dict[str, Any]:
        try:
            with open(filename, encoding="utf8") as cfile:
                loaded_config = load(cfile, Loader=Loader)
        except FileNotFoundError:
            print("No config found, using default values")
            loaded_config = {}

        return self.complete_config(loaded_config)

    def update_config(self, config: dict[str, Any]) -> None:
        self.general_config.load_config(config["config"])
        for source_name, source_config in config["sources"].items():
            self.tabs[source_name].load_config(source_config)

        self.update_qr()

    def save_config(self) -> None:
        os.makedirs(os.path.dirname(self.configfile), exist_ok=True)

        with open(self.configfile, "w", encoding="utf-8") as f:
            dump(self.gather_config(), f, Dumper=Dumper)

    def gather_config(self) -> dict[str, Any]:
        sources = {}
        for source, tab in self.tabs.items():
            sources[source] = tab.get_config()

        general_config = self.general_config.get_config() | {
            "show_advanced": self.show_advanced_toggle.isChecked()
        }

        return {"sources": sources, "config": general_config}

    def import_config(self) -> None:
        filename = QFileDialog.getOpenFileName(self, "Open File", "", "YAML Files (*.yaml)")[0]

        if filename:
            config = self.load_config(filename)
            self.update_config(config)

    def export_config(self) -> None:
        filename = QFileDialog.getSaveFileName(self, "Save File", "", "YAML Files (*.yaml)")[0]
        if filename:
            config = self.gather_config()

            with open(filename, "w", encoding="utf-8") as f:
                dump(config, f, Dumper=Dumper)

    def check_if_client_is_running(self) -> None:
        if self.client is None:
            self.timer.stop()
            return

        if not self.client.connection_state.is_connected():
            self.client = None
            self.set_client_button_start()
        else:
            self.set_client_button_stop()

    def set_client_button_stop(self) -> None:
        self.general_config.string_options["server"].setEnabled(False)
        self.general_config.string_options["room"].setEnabled(False)
        self.update_config_button.setDisabled(False)
        self.remove_room_button.setDisabled(False)
        self.export_queue_button.setDisabled(False)
        self.import_queue_button.setDisabled(False)

        self.startbutton.setText("Disconnect")

    def set_client_button_start(self) -> None:
        self.general_config.string_options["server"].setEnabled(True)
        self.general_config.string_options["room"].setEnabled(True)
        self.update_config_button.setDisabled(True)
        self.remove_room_button.setDisabled(True)
        self.export_queue_button.setDisabled(True)
        self.import_queue_button.setDisabled(True)

        self.startbutton.setText("Connect")

    def start_syng_client(self) -> None:
        if self.client is None or not self.client.connection_state.is_connected():
            logger.debug("Starting client")
            self.save_config()
            config = self.gather_config()
            self.client = Client(config)
            asyncio.run_coroutine_threadsafe(self.client.start_client(config), self.loop)
            # model = QueueModel(self.client.state.queue)
            # self.queue_list_view.setModel(model)
            # self.client.add_queue_callback(model.update)
            self.timer.start(500)
            self.set_client_button_stop()
        else:
            logger.debug("Stopping client")
            self.client.quit_callback()
            self.set_client_button_start()

    @pyqtSlot(str, int)
    def print_log(self, log: str, level: int) -> None:
        if level == logging.CRITICAL:
            log_msg_box = QMessageBox(self)
            log_msg_box.setIcon(QMessageBox.Icon.Critical)
            log_msg_box.setWindowTitle("Critical Error")
            log_msg_box.setText(log)
            log_msg_box.exec()
        self.log_text.append(f"[{datetime.now().strftime('%H:%M:%S')}] {log}")

    def change_qr(self, data: str) -> None:
        qr = QRCode(box_size=10, border=2)
        qr.add_data(data)
        qr.make()
        image = qr.make_image().convert("RGB")
        buf = BytesIO()
        image.save(buf, "PNG")
        qr_pixmap = QPixmap()
        qr_pixmap.loadFromData(buf.getvalue(), "PNG")
        self.qr_label.setPixmap(qr_pixmap)

    def update_qr(self) -> None:
        config = self.general_config.get_config()
        syng_server = config["server"]
        syng_server += "" if syng_server.endswith("/") else "/"
        room = config["room"]
        self.linklabel.setText(
            f'<center><a href="{syng_server + room}">{syng_server + room}</a><center>'
        )
        self.linklabel.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.change_qr(syng_server + room)


class LoggingLabelHandler(logging.Handler):
    class LogSignalEmiter(QObject):
        log_signal = pyqtSignal(str, int)

        def __init__(self, parent: Optional[QObject] = None):
            super().__init__(parent)

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__()
        self.log_signal_emiter = self.LogSignalEmiter(parent)
        self._cleanup = False

    def emit(self, record: logging.LogRecord) -> None:
        if not self._cleanup:  # This could race condition, but it's not a big
            # deal since it only causes a race condition,
            # when the program ends
            self.log_signal_emiter.log_signal.emit(self.format(record), record.levelno)

    def cleanup(self) -> None:
        self._cleanup = True


def run_gui() -> None:
    os.makedirs(platformdirs.user_cache_dir("syng"), exist_ok=True)
    base_dir = os.path.dirname(__file__)
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        base_dir = getattr(sys, "_MEIPASS")

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QApplication([])
    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    if os.name == "nt":
        app.setWindowIcon(QIcon(os.path.join(base_dir, "syng.ico")))
    else:
        app.setWindowIcon(QIcon(":/icons/syng.ico"))
    app.setApplicationName("Syng.Rocks!")
    app.setDesktopFileName("rocks.syng.Syng")
    window = SyngGui()
    window.show()
    with event_loop:
        event_loop.run_forever()


if __name__ == "__main__":
    run_gui()
